import numpy as np
import logging

from openfisca_core import periods
from openfisca_core.taxscales import MarginalRateTaxScale, combine_tax_scales
from openfisca_core.formula_helpers import switch
from openfisca_france.model.base import TypesCategorieSalarie, TAUX_DE_PRIME
from openfisca_france.model.prelevements_obligatoires.prelevements_sociaux.cotisations_sociales.base import (
    cotisations_salarie_by_categorie_salarie,
    )
from openfisca_france_data.smic import smic_horaire_brut


log = logging.getLogger(__name__)


def get_baremes_salarie(parameters, categorie_salarie, period, exclude_alsace_moselle = True):
    assert categorie_salarie in [
        "prive_cadre",
        "prive_non_cadre",
        "public_non_titulaire",
        "public_titulaire_etat",
        "public_titulaire_hospitaliere",
        "public_titulaire_territoriale",
        ]

    cotisations_salaries = cotisations_salarie_by_categorie_salarie[categorie_salarie]
    cotisations_salarie_retenues = list()

    for cotisation_salarie in cotisations_salaries:
        bareme = parameters.cotsoc.cotisations_salarie.children[categorie_salarie].children[cotisation_salarie]
        if exclude_alsace_moselle and "alsace_moselle" in cotisation_salarie:
            continue

        if bareme(period).thresholds and bareme(period).rates:
            cotisations_salarie_retenues.append(cotisation_salarie)

    return set(cotisations_salarie_retenues)



def create_salaire_de_base(individus, period = None, revenu_type = 'imposable', tax_benefit_system = None):
    """
    Calcule la variable salaire_de_base à partir du salaire imposable par inversion du barème
    de cotisations sociales correspondant à la catégorie à laquelle appartient le salarié.
    """
    assert period is not None
    assert revenu_type in ['net', 'imposable']
    for variable in ['categorie_salarie', 'contrat_de_travail', 'heures_remunerees_volume']:
        assert variable in individus.columns, "{} is missing".format(variable)
    assert tax_benefit_system is not None

    parameters = tax_benefit_system.parameters(period.start)
    salarie = parameters.cotsoc.cotisations_salarie
    plafond_securite_sociale_mensuel = parameters.prelevements_sociaux.pss.plafond_securite_sociale_mensuel
    parameters_csg_deductible = parameters.prelevements_sociaux.contributions_sociales.csg.activite.deductible
    taux_csg = parameters_csg_deductible.taux
    taux_abattement = parameters_csg_deductible.abattement.rates[0]
    try:
        seuil_abattement = parameters_csg_deductible.abattement.thresholds[1]
    except IndexError:  # Pour gérer le fait que l'abattement n'a pas toujours était limité à 4 PSS
        seuil_abattement = None
    csg_deductible = MarginalRateTaxScale(name = 'csg_deductible')
    csg_deductible.add_bracket(0, taux_csg * (1 - taux_abattement))
    if seuil_abattement is not None:
        csg_deductible.add_bracket(seuil_abattement, taux_csg)

    if revenu_type == 'net':  # On ajoute CSG imposable et crds

        parameters_csg_imposable = parameters.prelevements_sociaux.contributions_sociales.csg.activite.imposable
        taux_csg = parameters_csg_imposable.taux
        taux_abattement = parameters_csg_imposable.abattement.rates[0]
        try:
            seuil_abattement = parameters_csg_imposable.abattement.thresholds[1]
        except IndexError:  # Pour gérer le fait que l'abattement n'a pas toujours était limité à 4 PSS
            seuil_abattement = None
        csg_imposable = MarginalRateTaxScale(name = 'csg_imposable')
        csg_imposable.add_bracket(0, taux_csg * (1 - taux_abattement))
        if seuil_abattement is not None:
            csg_imposable.add_bracket(seuil_abattement, taux_csg)

        parameters_crds = parameters.prelevements_sociaux.contributions_sociales.crds.activite
        taux_csg = parameters_crds.taux
        taux_abattement = parameters_crds.abattement.rates[0]
        try:
            seuil_abattement = parameters_crds.abattement.thresholds[1]
        except IndexError:  # Pour gérer le fait que l'abattement n'a pas toujours était limité à 4 PSS
            seuil_abattement = None
        crds = MarginalRateTaxScale(name = 'crds')
        crds.add_bracket(0, taux_csg * (1 - taux_abattement))
        if seuil_abattement is not None:
            crds.add_bracket(seuil_abattement, taux_csg)

    # Check baremes
    target_by_categorie_salarie = dict(
        (categorie_salarie, get_baremes_salarie(tax_benefit_system.parameters, categorie_salarie, period))
        for categorie_salarie in ['prive_cadre', 'prive_non_cadre', 'public_non_titulaire']
        )

    for categorie_salarie, target in target_by_categorie_salarie.items():
        baremes_collection = salarie[categorie_salarie]
        baremes_to_remove = list()
        for name, bareme in baremes_collection._children.items():
            if name not in target:
                baremes_to_remove.append(name)

        # We split since we cannot remove from dict while iterating
        for name in baremes_to_remove:
            del baremes_collection._children[name]

    for categorie_salarie, target in target_by_categorie_salarie.items():
        test = set(
            name for name, bareme in salarie[categorie_salarie]._children.items()
            # if isinstance(bareme, MarginalRateTaxScale)
            )
        # assert target[categorie] == test, 'target: {} \n test {}'.format(target[categorie], test)
    del bareme

    # On ajoute la CSG deductible et on proratise par le plafond de la sécurité sociale
    # Pour éviter les divisions 0 /0 dans le switch qui sert à calculer le salaire_pour_inversion_proratise
    whours = parameters.marche_travail.salaire_minimum.smic.nb_heures_travail_mensuel

    if period.unit == 'year':
        plafond_securite_sociale = plafond_securite_sociale_mensuel * 12
        heures_temps_plein = whours * 12
    elif period.unit == 'month':
        plafond_securite_sociale = plafond_securite_sociale_mensuel * period.size
        heures_temps_plein = whours * period.size
    else:
        raise

    if revenu_type == 'imposable':
        salaire_pour_inversion = individus.salaire_imposable
    else:
        salaire_pour_inversion = individus.salaire_net

    categorie_salarie = individus.categorie_salarie
    contrat_de_travail = individus.contrat_de_travail
    heures_remunerees_volume = individus.heures_remunerees_volume

    heures_remunerees_volume_avoid_warning = heures_remunerees_volume + (heures_remunerees_volume == 0) * 1e9
    salaire_pour_inversion_proratise = switch(
        contrat_de_travail,
        {
            # temps plein
            0: salaire_pour_inversion / plafond_securite_sociale,
            # temps partiel
            1: salaire_pour_inversion / (
                (heures_remunerees_volume_avoid_warning / heures_temps_plein) * plafond_securite_sociale
                ),
            }
        )

    def add_agirc_gmp_to_agirc(agirc, parameters):
        plafond_securite_sociale_annuel = parameters.prelevements_sociaux.pss.plafond_securite_sociale_mensuel * 12
        salaire_charniere = parameters.prelevements_sociaux.regimes_complementaires_retraite_secteur_prive.gmp.salaire_charniere_annuel / plafond_securite_sociale_annuel
        cotisation = parameters.prelevements_sociaux.regimes_complementaires_retraite_secteur_prive.gmp.cotisation_forfaitaire_mensuelle.part_salariale * 12
        n = (cotisation + 1) * 12
        agirc.add_bracket(n / plafond_securite_sociale_annuel, 0)
        agirc.rates[0] = cotisation / n
        agirc.thresholds[2] = salaire_charniere

    salaire_de_base = 0.0
    for categorie in ['prive_non_cadre', 'prive_cadre', 'public_non_titulaire']:
        if categorie == 'prive_cadre' and "agirc" in salarie[categorie]._children:
            print("adding GMP")
            add_agirc_gmp_to_agirc(salarie[categorie].agirc, parameters)

        bareme = combine_tax_scales(salarie[categorie])
        bareme.add_tax_scale(csg_deductible)
        if revenu_type == 'net':
            bareme.add_tax_scale(csg_imposable)
            bareme.add_tax_scale(crds)

        assert bareme.inverse().thresholds[0] == 0, "Invalid inverse bareme for {}:\n {}".format(
            categorie, bareme.inverse())
        for rate in bareme.inverse().rates:  # Vérification que l'on peut inverser correctement
            assert rate > 0

        brut_proratise = bareme.inverse().calc(salaire_pour_inversion_proratise)
        assert np.isfinite(brut_proratise).all()
        brut = plafond_securite_sociale * switch(
            contrat_de_travail,
            {
                # temps plein
                0: brut_proratise,
                # temps partiel
                1: brut_proratise * (heures_remunerees_volume / (heures_temps_plein)),
                }
            )
        salaire_de_base += (
            (categorie_salarie == TypesCategorieSalarie[categorie].index) * brut
            )

    individus['salaire_de_base'] = salaire_de_base

    if period.unit == 'year':
        year = period.start.year
        individus.loc[
            (individus.contrat_de_travail == 0) & (individus.salaire_de_base > 0),
            'salaire_de_base'
            ] = np.maximum(salaire_de_base, smic_horaire_brut[year] * 35 * 52 / 12)
        individus.loc[
            (individus.contrat_de_travail == 1) & (individus.salaire_de_base > 0),
            'salaire_de_base'
            ] = np.maximum(salaire_de_base, smic_horaire_brut[year] * heures_remunerees_volume)


def create_traitement_indiciaire_brut(individus, period = None, revenu_type = 'imposable',
            tax_benefit_system = None):
    """
    Calcule le tratement indiciaire brut à partir du salaire imposable ou du salaire net.
    Note : le supplément familial de traitement est imposable. Pas géré
    """
    assert period is not None
    assert revenu_type in ['net', 'imposable']
    assert tax_benefit_system is not None

    for variable in ['categorie_salarie', 'contrat_de_travail', 'heures_remunerees_volume']:
        assert variable in individus.columns

    if revenu_type == 'imposable':
        assert 'salaire_imposable' in individus.columns
        salaire_pour_inversion = individus.salaire_imposable
    else:
        assert 'salaire_net' in individus.columns
        salaire_pour_inversion = individus.salaire_net

    categorie_salarie = individus.categorie_salarie
    contrat_de_travail = individus.contrat_de_travail
    heures_remunerees_volume = individus.heures_remunerees_volume

    parameters = tax_benefit_system.parameters(period.start)

    salarie = parameters.cotsoc.cotisations_salarie
    plafond_securite_sociale_mensuel = parameters.prelevements_sociaux.pss.plafond_securite_sociale_mensuel
    parameters_csg_deductible = parameters.prelevements_sociaux.contributions_sociales.csg.activite.deductible
    taux_csg = parameters_csg_deductible.taux
    taux_abattement = parameters_csg_deductible.abattement.rates[0]
    try:
        seuil_abattement = parameters_csg_deductible.abattement.thresholds[1]
    except IndexError:  # Pour gérer le fait que l'abattement n'a pas toujours été limité à 4 PSS
        seuil_abattement = None
    csg_deductible = MarginalRateTaxScale(name = 'csg_deductible')
    csg_deductible.add_bracket(0, taux_csg * (1 - taux_abattement))
    if seuil_abattement is not None:
        csg_deductible.add_bracket(seuil_abattement, taux_csg)

    if revenu_type == 'net':
        # Cas des revenus nets:
        # comme les salariés du privé, on ajoute CSG imposable et crds qui s'appliquent à tous les revenus
        # 1. csg imposable
        parameters_csg_imposable = parameters.prelevements_sociaux.contributions_sociales.csg.activite.imposable
        taux_csg = parameters_csg_imposable.taux
        taux_abattement = parameters_csg_imposable.abattement.rates[0]
        try:
            seuil_abattement = parameters_csg_imposable.abattement.thresholds[1]
        except IndexError:  # Pour gérer le fait que l'abattement n'a pas toujours été limité à 4 PSS
            seuil_abattement = None
        csg_imposable = MarginalRateTaxScale(name = 'csg_imposable')
        csg_imposable.add_bracket(0, taux_csg * (1 - taux_abattement))
        if seuil_abattement is not None:
            csg_imposable.add_bracket(seuil_abattement, taux_csg)
        # 2. crds
        parameters_crds = parameters.prelevements_sociaux.contributions_sociales.crds.activite
        taux_csg = parameters_crds.taux
        taux_abattement = parameters_crds.abattement.rates[0]
        try:
            seuil_abattement = parameters_crds.abattement.thresholds[1]
        except IndexError:  # Pour gérer le fait que l'abattement n'a pas toujours été limité à 4 PSS
            seuil_abattement = None
        crds = MarginalRateTaxScale(name = 'crds')
        crds.add_bracket(0, taux_csg * (1 - taux_abattement))
        if seuil_abattement is not None:
            crds.add_bracket(seuil_abattement, taux_csg)

    # Check baremes
    target = dict()
    target['public_titulaire_etat'] = set(['excep_solidarite', 'pension', 'rafp'])
    target['public_titulaire_hospitaliere'] = set(['excep_solidarite', 'cnracl_s_ti', 'rafp'])
    target['public_titulaire_territoriale'] = set(['excep_solidarite', 'cnracl_s_ti', 'rafp'])

    categories_salarie_du_public = [
        'public_titulaire_etat',
        'public_titulaire_hospitaliere',
        'public_titulaire_territoriale',
        ]

    for categorie in categories_salarie_du_public:
        baremes_collection = salarie[categorie]
        test = set(
            name for name, bareme in salarie[categorie]._children.items()
            if isinstance(bareme, MarginalRateTaxScale) and name != 'cnracl_s_nbi'
            )
        # assert target[categorie] == test, 'target for {}: \n  target = {} \n  test = {}'.format(categorie, target[categorie], test)

    # Barèmes à éliminer :
        # cnracl_s_ti = taux hors NBI -> OK
        # cnracl_s_nbi = taux NBI -> On ne le prend pas en compte pour l'instant
    for categorie in [
        'public_titulaire_hospitaliere',
        'public_titulaire_territoriale',
        ]:
        baremes_collection = salarie[categorie]
        baremes_to_remove = list()
        baremes_to_remove.append('cnracl_s_nbi')
        for name in baremes_to_remove:
            if 'cnracl_s_nbi' in baremes_collection:
                del baremes_collection._children[name]

    salarie = salarie._children.copy()
    # RAFP des agents titulaires
    for categorie in categories_salarie_du_public:
        baremes_collection = salarie[categorie]
        baremes_collection['rafp'].multiply_rates(TAUX_DE_PRIME, inplace = True)

    # On ajoute la CSG déductible et on proratise par le plafond de la sécurité sociale
    whours = parameters.marche_travail.salaire_minimum.smic.nb_heures_travail_mensuel

    if period.unit == 'year':
        plafond_securite_sociale = plafond_securite_sociale_mensuel * 12
        heures_temps_plein = whours * 12
    elif period.unit == 'month':
        plafond_securite_sociale = plafond_securite_sociale_mensuel * period.size
        heures_temps_plein = whours * period.size
    else:
        raise

    # Pour a fonction publique la csg est calculée sur l'ensemble salbrut(=TIB) + primes
    # Imposable = (1 + taux_prime) * TIB - csg[(1 + taux_prime) * TIB] - pension[TIB]
    for categorie in categories_salarie_du_public:
        bareme_csg_deduc_public = csg_deductible.multiply_rates(
            1 + TAUX_DE_PRIME, inplace = False, new_name = "csg deduc public")
        if revenu_type == 'net':
            bareme_csg_imp_public = csg_imposable.multiply_rates(
                1 + TAUX_DE_PRIME, inplace = False, new_name = "csg imposable public")
            bareme_crds_public = crds.multiply_rates(
                1 + TAUX_DE_PRIME, inplace = False, new_name = "crds public")

    for categorie in categories_salarie_du_public:
        bareme_prime = MarginalRateTaxScale(name = "taux de prime")
        bareme_prime.add_bracket(0, -TAUX_DE_PRIME)  # barème équivalent à taux_prime*TIB

    heures_remunerees_volume_avoid_warning = heures_remunerees_volume + (heures_remunerees_volume == 0) * 1e9
    salaire_pour_inversion_proratise = switch(
        contrat_de_travail,
        {
            # temps plein
            0: salaire_pour_inversion / plafond_securite_sociale,
            # temps partiel
            1: salaire_pour_inversion / (
                (heures_remunerees_volume_avoid_warning / heures_temps_plein) * plafond_securite_sociale
                ),
            }
        )

    traitement_indiciaire_brut = 0.0

    for categorie in categories_salarie_du_public:
        for key, value in salarie[categorie]._children.items():
            log.debug(key, value)
        bareme = combine_tax_scales(salarie[categorie])
        log.debug('bareme cotsoc : {}'.format(bareme))
        bareme.add_tax_scale(bareme_csg_deduc_public)
        log.debug('bareme cotsoc + csg_deduc: {}'.format(bareme))
        if revenu_type == 'net':
            bareme.add_tax_scale(bareme_csg_imp_public)
            log.debug('bareme cotsoc + csg_deduc + csg_imp: {}'.format(bareme))
            bareme.add_tax_scale(bareme_crds_public)
            log.debug('bareme cotsoc + csg_deduc + csg_imp + crds: {}'.format(bareme))

        bareme.add_tax_scale(bareme_prime)
        log.debug('bareme cotsoc + csg_deduc + csg_imp + crds + prime: {}'.format(bareme))

        brut_proratise = bareme.inverse().calc(salaire_pour_inversion_proratise)

        assert np.isfinite(brut_proratise).all()
        brut = plafond_securite_sociale * switch(
            contrat_de_travail,
            {
                # temps plein
                0: brut_proratise,
                # temps partiel
                1: brut_proratise * (heures_remunerees_volume / (heures_temps_plein)),
                }
            )
        traitement_indiciaire_brut += (
            (categorie_salarie == TypesCategorieSalarie[categorie].index) * brut
            )
        if (categorie_salarie == TypesCategorieSalarie[categorie].index).any():
            log.debug("Pour {} : brut = {}".format(TypesCategorieSalarie[categorie].index, brut))
            log.debug('bareme direct: {}'.format(bareme))

    individus['traitement_indiciaire_brut'] = traitement_indiciaire_brut
    individus['primes_fonction_publique'] = TAUX_DE_PRIME * traitement_indiciaire_brut


def create_revenus_remplacement_bruts(individus, period, tax_benefit_system):
    assert 'taux_csg_remplacement' in individus

    individus.chomage_imposable.fillna(0, inplace = True)
    individus.retraite_imposable.fillna(0, inplace = True)

    parameters = tax_benefit_system.get_parameters_at_instant(period.start)
    csg = parameters.prelevements_sociaux.contributions_sociales.csg
    csg_deductible_chomage = csg.remplacement.allocations_chomage.deductible
    taux_plein = csg_deductible_chomage.taux_plein
    taux_reduit = csg_deductible_chomage.taux_reduit
    seuil_chomage_net_exoneration = (
        (35 * 52) * smic_horaire_brut[period.start.year]
        * (
            (individus.taux_csg_remplacement == 2) / (1 - taux_reduit)
            + (individus.taux_csg_remplacement >= 3) / (1 - taux_plein)
            )
        )
    exonere_csg_chomage = (
        (individus.taux_csg_remplacement < 2)
        | (individus.chomage_imposable <= seuil_chomage_net_exoneration)
        )
    individus['chomage_brut'] = np.where(
        exonere_csg_chomage,
        individus.chomage_imposable,
        (individus.taux_csg_remplacement == 2) * individus.chomage_imposable / (1 - taux_reduit)
        + (individus.taux_csg_remplacement >= 3) * individus.chomage_imposable / (1 - taux_plein)
        )
    assert individus['chomage_brut'].notnull().all()

    csg_deductible_retraite = parameters.prelevements_sociaux.contributions_sociales.csg.remplacement.pensions_retraite_invalidite.deductible
    taux_plein = csg_deductible_retraite.taux_plein
    taux_reduit = csg_deductible_retraite.taux_reduit
    if period.start.year >= 2019:
        taux_median = csg_deductible_retraite.taux_median
        individus['retraite_brute'] = (
        (individus.taux_csg_remplacement < 2) * individus.retraite_imposable
        + (individus.taux_csg_remplacement == 2) * individus.retraite_imposable / (1 - taux_reduit)
        + (individus.taux_csg_remplacement == 3) * individus.retraite_imposable / (1 - taux_median)
        + (individus.taux_csg_remplacement == 4) * individus.retraite_imposable / (1 - taux_plein)
        )
    else:
        individus['retraite_brute'] = (
            (individus.taux_csg_remplacement < 2) * individus.retraite_imposable
            + (individus.taux_csg_remplacement == 2) * individus.retraite_imposable / (1 - taux_reduit)
            + (individus.taux_csg_remplacement >= 3) * individus.retraite_imposable / (1 - taux_plein)
            )
    assert individus['retraite_brute'].notnull().all()



#! /usr/bin/env python
import gc
import logging
from pandas import concat, DataFrame


from openfisca_survey_manager.temporary import temporary_store_decorator
from openfisca_france_data.utils import assert_dtype

log = logging.getLogger(__name__)

# Retreives the families
# Creates 'idfam' and 'quifam' variables

AGE_RSA = 25

def control_04(dataframe, base):
    log.info("longueur de la dataframe après opération :".format(len(dataframe.index)))
    log.info("contrôle des doublons : il y a {} individus en double".format(
        any(dataframe.duplicated(subset = 'noindiv'))))
    log.info("contrôle des colonnes : il y a {} colonnes".format(len(dataframe.columns)))
    log.info("Il y a {} de familles différentes".format(len(dataframe.noifam.unique())))
    log.info("contrôle: {} noifam are NaN:".format(len(dataframe[dataframe['noifam'].isnull()])))
    log.info("{} lignes dans dataframe vs {} lignes dans base".format(len(dataframe.index), len(base.index)))
    assert len(dataframe.index) <= len(base.index), "dataframe has too many rows compared to base"
    assert set(dataframe.noifam.unique()).issubset(set(base.noindiv)), \
        "The following noindiv are not in the dataframe: \n {}".format(
            dataframe.loc[~dataframe.noifam.isin(base.noindiv), ['noifam', 'famille']]
        )


def subset_base(base, famille):
    """
    Generates a dataframe containing the values of base that are not already in famille
    """
    return base[~(base.noindiv.isin(famille.noindiv.values))].copy()


@temporary_store_decorator(file_name = 'erfs')
def famille(temporary_store = None, year = None):
    assert temporary_store is not None
    assert year is not None

    log.info('step_04_famille: construction de la table famille')

    # On suit la méthode décrite dans le Guide ERF_2002_rétropolée page 135
    # TODO: extraire ces valeurs d'un fichier de paramètres de législation
    if year == 2006:
        smic = 1254
    elif year == 2007:
        smic = 1280
    elif year == 2008:
        smic = 1308
    elif year == 2009:
        smic = 1337
    else:
        log.info("smic non défini")

    # TODO check if we can remove acteu forter etc since dealt with in 01_pre_proc

    log.info('Etape 1 : préparation de base')
    log.info('    1.1 : récupération de indivi')
    indivi = temporary_store['indivim_{}'.format(year)]

    indivi['year'] = year
    indivi["noidec"] = indivi["declar1"].str[0:2].copy()  # Not converted to int because some NaN are present
    indivi["agepf"] = (
        (indivi.naim < 7) * (indivi.year - indivi.naia) +
        (indivi.naim >= 7) * (indivi.year - indivi.naia - 1)
        ).astype(object)  # TODO: naia has some NaN but naim do not and then should be an int

    indivi = indivi[~(
        (indivi.lien == 6) & (indivi.agepf < 16) & (indivi.quelfic == "EE")
        )].copy()

    assert_dtype(indivi.year, "int64")
    assert_dtype(indivi.agepf, "int64")

    for series_name in ['agepf', 'noidec']:  # integer with NaN
        assert_dtype(indivi[series_name], "object")

    log.info('    1.2 : récupération des enfants à naître')
    individual_variables = [
        'acteu',
        'actrec',
        'agepf',
        'agepr',
        'cohab',
        'contra',
        'declar1',
        'forter',
        'ident',
        'lien',
        'lpr',
        'mrec',
        'naia',
        'naim',
        'noi',
        'noicon',
        'noidec',
        'noimer',
        'noindiv',
        'noiper',
        'persfip',
        'quelfic',
        'retrai',
        'rga',
        'rstg',
        'sexe',
        'stc',
        'titc',
        'year',
        'ztsai',
        ]
    enfants_a_naitre = temporary_store['enfants_a_naitre_{}'.format(year)][individual_variables].copy()
    enfants_a_naitre.drop_duplicates('noindiv', inplace = True)
    log.info(""""
    Il y a {} enfants à naitre avant de retirer ceux qui ne sont pas enfants
    de la personne de référence
    """.format(len(enfants_a_naitre.index)))
    enfants_a_naitre = enfants_a_naitre[enfants_a_naitre.lpr == 3].copy()
    enfants_a_naitre = enfants_a_naitre[~(enfants_a_naitre.noindiv.isin(indivi.noindiv.values))].copy()
    log.info(""""
    Il y a {} enfants à naitre après avoir retiré ceux qui ne sont pas enfants
    de la personne de référence
    """.format(len(enfants_a_naitre.index)))
    # PB with vars "agepf"  "noidec" "year"  NOTE: quels problèmes ? JS
    log.info("    1.3 : création de la base complète")
    base = concat([indivi, enfants_a_naitre])
    log.info("base contient {} lignes ".format(len(base.index)))
    base['noindiv'] = (100 * base.ident + base['noi']).astype(int)
    base['moins_de_15_ans_inclus'] = base.agepf < 16
    base['jeune_non_eligible_rsa'] = (base.agepf >= 16) & (base.agepf < AGE_RSA)
    base['jeune_eligible_rsa'] = base.agepf >= AGE_RSA
    base['ztsai'].fillna(0, inplace = True)
    base['smic55'] = base['ztsai'] >= (smic * 12 * 0.55)  # 55% du smic mensuel brut
    base['famille'] = 0
    base['kid'] = False
    for series_name in ['kid', 'moins_de_15_ans_inclus', 'jeune_non_eligible_rsa', 'jeune_eligible_rsa', 'smic55']:
        assert_dtype(base[series_name], "bool")
    assert_dtype(base.famille, "int")
    # TODO: remove or clean from NA assert_dtype(base.ztsai, "int")

    log.info("Etape 2 : On cherche les enfants ayant père et/ou mère")
    personne_de_reference = base[['ident', 'noi']][base.lpr == 1].copy()
    personne_de_reference['noifam'] = (100 * personne_de_reference.ident + personne_de_reference['noi']).astype(int)
    personne_de_reference = personne_de_reference[['ident', 'noifam']].copy()
    log.info("length personne_de_reference : {}".format(len(personne_de_reference.index)))
    nof01 = base[(base.lpr.isin([1, 2])) | ((base.lpr == 3) & (base.moins_de_15_ans_inclus)) |
                 ((base.lpr == 3) & (base.jeune_non_eligible_rsa) & (~base.smic55))].copy()
    log.info('longueur de nof01 avant merge : {}'.format(len(nof01.index)))
    nof01 = nof01.merge(personne_de_reference, on='ident', how='outer')
    nof01['famille'] = 10
    nof01['kid'] = (
        (nof01.lpr == 3) & (nof01.moins_de_15_ans_inclus)
        ) | (
            (nof01.lpr == 3) & (nof01.jeune_non_eligible_rsa) & ~(nof01.smic55)
            )
    for series_name in ['famille', 'noifam']:
        assert_dtype(nof01[series_name], "int")
    assert_dtype(nof01.kid, "bool")
    famille = nof01.copy()
    del nof01
    control_04(famille, base)

    log.info("    2.1 : identification des couples")
    # l'ID est le noi de l'homme

    hcouple = subset_base(base, famille)
    hcouple = hcouple[(hcouple.cohab == 1) & (hcouple.lpr >= 3) & (hcouple.sexe == 1)].copy()
    hcouple['noifam'] = (100 * hcouple.ident + hcouple.noi).astype(int)
    hcouple['famille'] = 21
    for series_name in ['famille', 'noifam']:
        assert_dtype(hcouple[series_name], "int")
    log.info("longueur hcouple : ".format(len(hcouple.index)))

    log.info("    2.2 : attributing the noifam to the wives")
    fcouple = base[~(base.noindiv.isin(famille.noindiv.values))].copy()
    fcouple = fcouple[(fcouple.cohab == 1) & (fcouple.lpr >= 3) & (fcouple.sexe == 2)].copy()
    # l'identifiant de la famille est celui du conjoint de la personne de référence du ménage
    fcouple['noifam'] = (100 * fcouple.ident + fcouple.noicon).astype(int)
    fcouple['famille'] = 22
    for series_name in ['famille', 'noifam']:
        assert_dtype(fcouple[series_name], "int")
    log.info("Il y a {} enfants avec parents en fcouple".format(len(fcouple.index)))

    famcom = fcouple.merge(hcouple, on='noifam', how='outer')
    log.info("longueur fancom après fusion : {}".format(len(famcom.index)))
    fcouple = fcouple.merge(famcom)  # TODO : check s'il ne faut pas faire un inner merge sinon présence de doublons
    log.info("longueur fcouple après fusion : {}".format(len(fcouple.index)))
    famille = concat([famille, hcouple, fcouple], join='inner')
    control_04(famille, base)

    log.info("Etape 3: Récupération des personnes seules")
    log.info("    3.1 : personnes seules de catégorie 1")
    seul1 = base[~(base.noindiv.isin(famille.noindiv.values))].copy()
    seul1 = seul1[(seul1.lpr.isin([3, 4])) & ((seul1.jeune_non_eligible_rsa & seul1.smic55) | seul1.jeune_eligible_rsa) & (seul1.cohab == 1) &
                  (seul1.sexe == 2)].copy()
    if len(seul1.index) > 0:
        seul1['noifam'] = (100 * seul1.ident + seul1.noi).astype(int)
        seul1['famille'] = 31
        for series_name in ['famille', 'noifam']:
            assert_dtype(seul1[series_name], "int")
        famille = concat([famille, seul1])
    control_04(famille, base)

    log.info("    3.1 personnes seules de catégorie 2")
    seul2 = base[~(base.noindiv.isin(famille.noindiv.values))].copy()
    seul2 = seul2[(seul2.lpr.isin([3, 4])) & seul2.jeune_non_eligible_rsa & seul2.smic55 & (seul2.cohab != 1)].copy()
    seul2['noifam'] = (100 * seul2.ident + seul2.noi).astype(int)
    seul2['famille'] = 32
    for series_name in ['famille', 'noifam']:
        assert_dtype(seul2[series_name], "int")
    famille = concat([famille, seul2])
    control_04(famille, base)

    log.info("    3.3 personnes seules de catégorie 3")
    seul3 = subset_base(base, famille)
    seul3 = seul3[(seul3.lpr.isin([3, 4])) & seul3.jeune_eligible_rsa & (seul3.cohab != 1)].copy()
    # TODO: CHECK erreur dans le guide méthodologique ERF 2002 lpr 3,4 au lieu de 3 seulement
    seul3['noifam'] = (100 * seul3.ident + seul3.noi).astype(int)
    seul3['famille'] = 33
    for series_name in ['famille', 'noifam']:
        assert_dtype(seul3[series_name], "int")
    famille = concat([famille, seul3])
    control_04(famille, base)

    log.info("    3.4 : personnes seules de catégorie 4")
    seul4 = subset_base(base, famille)
    assert seul4.noimer.notnull().all()
    seul4 = seul4[(seul4.lpr == 4) & seul4.jeune_non_eligible_rsa & ~(seul4.smic55) & (seul4.noimer == 0) &
                  (seul4.persfip == 'vous')].copy()
    if len(seul4.index) > 0:
        seul4['noifam'] = (100 * seul4.ident + seul4.noi).astype(int)
        seul4['famille'] = 34
        famille = concat([famille, seul4])
        for series_name in ['famille', 'noifam']:
            assert_dtype(seul4[series_name], "int")
    control_04(famille, base)

    log.info("Etape 4 : traitement des enfants")
    log.info("    4.1 : enfant avec mère")
    avec_mere = subset_base(base, famille)
    avec_mere = avec_mere[((avec_mere.lpr == 4) & ((avec_mere.jeune_non_eligible_rsa == 1) | (avec_mere.moins_de_15_ans_inclus == 1)) &
                           (avec_mere.noimer > 0))].copy()
    avec_mere['noifam'] = (100 * avec_mere.ident + avec_mere.noimer).astype(int)
    avec_mere['famille'] = 41
    avec_mere['kid'] = True
    for series_name in ['famille', 'noifam']:
        assert_dtype(avec_mere[series_name], "int")
    assert_dtype(avec_mere.kid, "bool")

    # On récupère les mères des enfants
    mereid = DataFrame(avec_mere['noifam'].copy())  # Keep a DataFrame instead of a Series to deal with rename and merge
    # Ces mères peuvent avoir plusieurs enfants, or il faut unicité de l'identifiant
    mereid.rename(columns = {'noifam': 'noindiv'}, inplace = True)
    mereid.drop_duplicates(inplace = True)
    mere = mereid.merge(base)
    mere['noifam'] = (100 * mere.ident + mere.noi).astype(int)
    mere['famille'] = 42
    for series_name in ['famille', 'noifam']:
        assert_dtype(mere[series_name], "int")
    avec_mere = avec_mere[avec_mere.noifam.isin(mereid.noindiv.values)].copy()
    log.info("Contrôle de famille après ajout des pères")
    control_04(mere, base)

    famille = famille[~(famille.noindiv.isin(mere.noindiv.values))].copy()
    control_04(famille, base)
    # on retrouve les conjoints des mères
    assert mere.noicon.notnull().all()
    conj_mereid = mere.loc[mere.noicon > 0, ['ident', 'noicon', 'noifam']].copy()
    conj_mereid['noindiv'] = 100 * conj_mereid.ident + conj_mereid.noicon
    assert_dtype(conj_mereid[series_name], "int")
    conj_mereid = conj_mereid[['noindiv', 'noifam']].copy()
    conj_mereid = conj_mereid.merge(base)
    control_04(conj_mereid, base)
    conj_mere = conj_mereid.merge(base)
    conj_mere['famille'] = 43
    for series_name in ['famille', 'noifam']:
        assert_dtype(conj_mereid[series_name], "int")
    famille = famille[~(famille.noindiv.isin(conj_mere.noindiv.values))].copy()

    famille = concat([famille, avec_mere, mere, conj_mere])
    control_04(famille, base)
    del avec_mere, mere, conj_mere, mereid, conj_mereid

    log.info("    4.2 : enfants avec père")
    avec_pere = subset_base(base, famille)
    assert avec_pere.noiper.notnull().all()
    avec_pere = avec_pere[(avec_pere.lpr == 4) &
                          ((avec_pere.jeune_non_eligible_rsa == 1) | (avec_pere.moins_de_15_ans_inclus == 1)) &
                          (avec_pere.noiper > 0)]
    avec_pere['noifam'] = (100 * avec_pere.ident + avec_pere.noiper).astype(int)
    avec_pere['famille'] = 44
    avec_pere['kid'] = True
    # TODO: hack to deal with the problem of presence of NaN in avec_pere
#    avec_pere.dropna(subset = ['noifam'], how = 'all', inplace = True)
    assert avec_pere['noifam'].notnull().all(), 'presence of NaN in avec_pere'
    for series_name in ['famille', 'noifam']:
        assert_dtype(avec_pere[series_name], "int")
    assert_dtype(avec_pere.kid, "bool")

    pereid = DataFrame(avec_pere['noifam'])  # Keep a DataFrame instead of a Series to deal with rename and merge
    pereid.rename(columns = {'noifam': 'noindiv'}, inplace = True)
    pereid.drop_duplicates(inplace = True)
    pere = pereid.merge(base)

    pere['noifam'] = (100 * pere.ident + pere.noi).astype(int)
    pere['famille'] = 45
    famille = famille[~(famille.noindiv.isin(pere.noindiv.values))].copy()

    # On récupère les conjoints des pères
    assert pere.noicon.notnull().all()
    conj_pereid = pere.loc[pere.noicon > 0, ['ident', 'noicon', 'noifam']].copy()
    conj_pereid['noindiv'] = (100 * conj_pereid.ident + conj_pereid.noicon).astype(int)
    conj_pereid = conj_pereid[['noindiv', 'noifam']].copy()

    conj_pere = conj_pereid.merge(base)
    control_04(conj_pere, base)
    if len(conj_pere.index) > 0:
        conj_pere['famille'] = 46
    for series_name in ['famille', 'noifam']:
        assert_dtype(conj_pere[series_name], "int")

    famille = famille[~(famille.noindiv.isin(conj_pere.noindiv.values))].copy()
    famille = concat([famille, avec_pere, pere, conj_pere])
    log.info("Contrôle de famille après ajout des pères")
    control_04(famille, base)
    del avec_pere, pere, pereid, conj_pere, conj_pereid

    log.info("    4.3 : enfants avec déclarant")
    avec_dec = subset_base(base, famille)
    avec_dec = avec_dec[
        (avec_dec.persfip == "pac") &
        (avec_dec.lpr == 4) &
        (
            (avec_dec.jeune_non_eligible_rsa & ~(avec_dec.smic55)) | (avec_dec.moins_de_15_ans_inclus == 1)
            )
        ]
    avec_dec['noifam'] = (100 * avec_dec.ident + avec_dec.noidec.astype('int')).astype('int')
    avec_dec['famille'] = 47
    avec_dec['kid'] = True
    for series_name in ['famille', 'noifam']:
        assert_dtype(avec_dec[series_name], "int")
    assert_dtype(avec_dec.kid, "bool")
    control_04(avec_dec, base)
    # on récupère les déclarants pour leur attribuer une famille propre
    declarant_id = DataFrame(avec_dec['noifam'].copy()).rename(columns={'noifam': 'noindiv'})
    declarant_id.drop_duplicates(inplace = True)
    dec = declarant_id.merge(base)
    dec['noifam'] = (100 * dec.ident + dec.noi).astype(int)
    dec['famille'] = 48
    for series_name in ['famille', 'noifam']:
        assert_dtype(dec[series_name], "int")
    famille = famille[~(famille.noindiv.isin(dec.noindiv.values))].copy()
    famille = concat([famille, avec_dec, dec])
    del dec, declarant_id, avec_dec
    control_04(famille, base)

    log.info("Etape 5 : Récupération des enfants fip")
    log.info("    5.1 : Création de la df fip")
    individual_variables_fip = [
        'acteu',
        'actrec',
        'agepf',
        'agepr',
        'cohab',
        'contra',
        'declar1',
        'forter',
        'ident',
        'lien',
        'lpr',
        'mrec',
        'naia',
        'naim',
        'noi',
        'noicon',
        'noidec',
        'noimer',
        'noindiv',
        'noiper',
        'persfip',
        'quelfic',
        'retrai',
        'rga',
        'rstg',
        'sexe',
        'stc',
        'titc',
        'year',
        'ztsai',
        ]
    fip = temporary_store['fipDat_{}'.format(year)][individual_variables_fip].copy()
    # Variables auxilaires présentes dans base qu'il faut rajouter aux fip'
    # WARNING les noindiv des fip sont construits sur les ident des déclarants
    # pas d'orvelap possible avec les autres noindiv car on a des noi =99, 98, 97 ,...'
    fip['moins_de_15_ans_inclus'] = (fip.agepf < 16)
    fip['jeune_non_eligible_rsa'] = ((fip.agepf >= 16) & (fip.agepf <= 20))
    fip['jeune_eligible_rsa'] = (fip.agepf >= 21)
    fip['smic55'] = (fip.ztsai >= smic * 12 * 0.55)
    fip['famille'] = 0
    fip['kid'] = False
    for series_name in ['kid', 'moins_de_15_ans_inclus', 'jeune_non_eligible_rsa', 'jeune_eligible_rsa', 'smic55']:
        assert_dtype(fip[series_name], "bool")
    for series_name in ['famille']:
        assert_dtype(fip[series_name], "int")

# # base <- rbind(base,fip)
# # table(base$quelfic)

# # enfant_fip <- base[(!base$noindiv %in% famille$noindiv),]
# # enfant_fip <- subset(enfant_fip, (quelfic=="FIP") & (( (agepf %in% c(19,20)) & !smic55 ) | (naia==year & rga=='6')) )  # TODO check year ou year-1 !
# # enfant_fip <- within(enfant_fip,{
# #                      noifam=100*ident+noidec
# #                      famille=50
# #                      kid=TRUE})
# # #                     ident=NA}) # TODO : je ne sais pas quoi mettre un NA fausse les manips suivantes
# # famille <- rbind(famille,enfant_fip)
# #
# # # TODO: En 2006 on peut faire ce qui suit car tous les parents fip sont déjà dans une famille
# # parent_fip <- famille[famille$noindiv %in% enfant_fip$noifam,]
# # any(enfant_fip$noifam %in% parent_fip$noindiv)
# # parent_fip <- within(parent_fip,{
# #                      noifam <- noindiv
# #                      famille <- 51
# #                      kid <- FALSE})
# # famille[famille$noindiv %in% enfant_fip$noifam,] <- parent_fip
# # # TODO quid du conjoint ?

    log.info("    5.2 : extension de base avec les fip")
    base_ = concat([base, fip])
    enfant_fip = subset_base(base_, famille)
    enfant_fip = enfant_fip[
        (enfant_fip.quelfic == "FIP") & (
            (enfant_fip.agepf.isin([19, 20]) & ~(enfant_fip.smic55)) |
            ((enfant_fip.naia == enfant_fip.year - 1) & (enfant_fip.rga.astype('int') == 6))
            )
        ].copy()
    enfant_fip['noifam'] = (100 * enfant_fip.ident + enfant_fip.noidec).astype(int)
    enfant_fip['famille'] = 50
    enfant_fip['kid'] = True
    enfant_fip['ident'] = None  # TODO: should we really do this ?
    assert_dtype(enfant_fip.kid, "bool")
    for series_name in ['famille', 'noifam']:
        assert_dtype(enfant_fip[series_name], "int")
    control_04(enfant_fip, base)
    famille = concat([famille, enfant_fip])
    base = concat([base, enfant_fip])
    parent_fip = famille[famille.noindiv.isin(enfant_fip.noifam.values)].copy()
    assert (enfant_fip.noifam.isin(parent_fip.noindiv.values)).any(), \
        "{} doublons entre enfant_fip et parent fip !".format((enfant_fip.noifam.isin(parent_fip.noindiv.values)).sum())
    parent_fip['noifam'] = parent_fip['noindiv'].values.copy()
    parent_fip['famille'] = 51
    parent_fip['kid'] = False
    log.info("Contrôle de parent_fip")
    control_04(parent_fip, base)
    control_04(famille, base)
    famille = famille.merge(parent_fip, how='outer')
    del enfant_fip, fip, parent_fip
    gc.collect()
    # duplicated_individuals = famille.noindiv.duplicated()
    # TODO: How to prevent failing in the next assert and avoiding droppping duplicates ?
    # assert not duplicated_individuals.any(), "{} duplicated individuals in famille".format(
    # duplicated_individuals.sum())
    famille = famille.drop_duplicates(subset = 'noindiv', take_last = True)
    control_04(famille, base)

# # message('Etape 6 : non attribué')
# # non_attribue1 <- base[(!base$noindiv %in% famille$noindiv),]
# # non_attribue1 <- subset(non_attribue1,
# #                         (quelfic!="FIP") & (moins_de_15_ans_inclus | (jeune_non_eligible_rsa&(lien %in% c(1,2,3,4) & agepr>=35)))
# #                         )
# # # On rattache les moins de 15 ans avec la PR (on a déjà éliminé les enfants en nourrice)
# # non_attribue1 <- merge(pr,non_attribue1)
# # non_attribue1 <- within(non_attribue1,{
# #   famille <- ifelse(moins_de_15_ans_inclus,61,62)
# #     kid <- TRUE })
# #
# # rm(pr)
# # famille <- rbind(famille,non_attribue1)
# # dup <- duplicated(famille$noindiv)
# # table(dup)
# # rm(non_attribue1)
# # table(famille$famille, useNA="ifany")
# #
# # non_attribue2 <- base[(!base$noindiv %in% famille$noindiv) & (base$quelfic!="FIP"),]
# # non_attribue2 <- within(non_attribue2,{
# #   noifam <- 100*ident+noi # l'identifiant est celui du jeune */
# #     kid<-FALSE
# #     famille<-63})
# #
# # famille <- rbind(famille,non_attribue2)

    log.info("Etape 6 : gestion des non attribués")
    log.info("    6.1 : non attribués type 1")
    non_attribue1 = subset_base(base, famille)
    non_attribue1 = non_attribue1[
        ~(non_attribue1.quelfic != 'FIP') & (
            non_attribue1.moins_de_15_ans_inclus | (
                non_attribue1.jeune_non_eligible_rsa & (non_attribue1.lien.isin(range(1, 5))) & (non_attribue1.agepr >= 35)
                )
            )
        ].copy()
    # On rattache les moins de 15 ans avec la PR (on a déjà éliminé les enfants en nourrice)
    non_attribue1 = non_attribue1.merge(personne_de_reference)
    control_04(non_attribue1, base)
    non_attribue1['famille'] = 61 * non_attribue1.moins_de_15_ans_inclus + 62 * ~(non_attribue1.moins_de_15_ans_inclus)
    non_attribue1['kid'] = True
    assert_dtype(non_attribue1.kid, "bool")
    assert_dtype(non_attribue1.famille, "int")
    famille = concat([famille, non_attribue1])
    control_04(famille, base)
    del personne_de_reference, non_attribue1

    log.info("    6.2 : non attribué type 2")
    non_attribue2 = base[(~(base.noindiv.isin(famille.noindiv.values)) & (base.quelfic != "FIP"))].copy()
    non_attribue2['noifam'] = (100 * non_attribue2.ident + non_attribue2.noi).astype(int)
    non_attribue2['kid'] = False
    non_attribue2['famille'] = 63
    assert_dtype(non_attribue2.kid, "bool")
    for series_name in ['famille', 'noifam']:
        assert_dtype(non_attribue2[series_name], "int")
    famille = concat([famille, non_attribue2], join='inner')
    control_04(famille, base)
    del non_attribue2

    # Sauvegarde de la table famille
    log.info("Etape 7 : Sauvegarde de la table famille")
    log.info("    7.1 : Mise en forme finale")
#    TODO: nettoyer les champs qui ne servent plus à rien
#    famille['idec'] = famille['declar1'].str[3:11]
#    famille['idec'].apply(lambda x: str(x)+'-')
#    famille['idec'] += famille['declar1'].str[0:2]
    famille['chef'] = (famille['noifam'] == (100 * famille.ident + famille.noi))
    assert_dtype(famille.chef, "bool")

    famille.reset_index(inplace = True)
    control_04(famille, base)

    log.info("    7.2 : création de la colonne rang")

    famille['rang'] = famille.kid.astype('int')
    while any(famille[(famille.rang != 0)].duplicated(subset = ['rang', 'noifam'])):
        famille.rang.loc[famille.rang != 0] += famille[famille.rang != 0].copy().duplicated(
            subset = ["rang", 'noifam']).values
        log.info("nb de rangs différents : {}".format(len(set(famille.rang.values))))

    log.info("    7.3 : création de la colonne quifam et troncature")
    log.info("value_counts chef : \n {}".format(famille['chef'].value_counts()))
    log.info("value_counts kid :' \n {}".format(famille['kid'].value_counts()))

    famille['quifam'] = -1
    # famille['quifam'] = famille['quifam'].where(famille['chef'].values, 0)
    # ATTENTTION : ^ stands for XOR
    famille.quifam = (0 +
        ((~famille['chef']) & (~famille['kid'])).astype(int) +
        famille.kid * famille.rang
        ).astype('int')

    # TODO: Test a groupby to improve the following this (should be placed )
    #    assert famille['chef'].sum() == len(famille.noifam.unique()), \
    #      'The number of family chiefs {} is different from the number of families {}'.format(
    #          famille['chef'].sum(),
    #          len(famille.idfam.unique())
    #          )

    #    famille['noifam'] = famille['noifam'].astype('int')
    log.info("value_counts quifam : \n {}".format(famille['quifam'].value_counts()))
    famille = famille[['noindiv', 'quifam', 'noifam']].copy()
    famille.rename(columns = {'noifam': 'idfam'}, inplace = True)
    log.info("Vérifications sur famille")
    # TODO: we drop duplicates if any

    duplicated_famillle_count = famille.duplicated(subset = ['idfam', 'quifam']).sum()
    if duplicated_famillle_count > 0:
        log.info("There are {} duplicates of quifam inside famille, we drop them".format(
            )
            )
        famille.drop_duplicates(subset = ['idfam', 'quifam'], inplace = True)
        # assert not(famille.duplicated(subset = ['idfam', 'quifam']).any()), \
        #   'There are {} duplicates of quifam inside famille'.format(
        #       famille.duplicated(subset = ['idfam', 'quifam']).sum())

    temporary_store["famc_{}".format(year)] = famille
    del indivi, enfants_a_naitre


if __name__ == '__main__':
    # import sys
    # logging.basicConfig(level = logging.INFO, stream = sys.stdout)
    logging.basicConfig(level = logging.INFO, filename = 'step_04.log', filemode = 'w')
    year = 2009
    famille(year = year)
    log.info("étape 04 famille terminée")



"""Compare openfisca-france-data simulation to erfs-fpr."""


import click
import logging


from openfisca_survey_manager.survey_collections import SurveyCollection
from openfisca_france_data.comparator import AbstractComparator
from openfisca_france_data.erfs_fpr import REFERENCE_YEAR
from openfisca_france_data.erfs_fpr.input_data_builder.step_01_preprocessing import build_table_by_name


log = logging.getLogger(__name__)


openfisca_by_erfs_fpr_variables = {
    "chomage_i": "chomage_imposable",
    "ident": "idmen_original",
    "noindiv": "noindiv",
    "rag_i": "rag",
    "retraites_i": "retraite_imposable",  # TODO: CHECk
    #"rev_fonciers_bruts": "f4ba",
    "ric_i": "ric",
    "rnc_i": "rnc",
    "salaires_i": "salaire_imposable",
    "logt": "statut_occupation_logement",
    "rev_fonciers_bruts": "revenu_categoriel_foncier_menage",
    "rev_valeurs_mobilieres_bruts": "revenus_capitaux_prelevement_forfaitaire_unique_ir_menage",
    "rev_financier_prelev_lib_imputes": "rev_financier_prelev_lib_imputes_menage",
    "ppa": "ppa_menage",
    "nivviem": "niveau_de_vie",
    "prest_fam_autres": "prestations_familiales_autres_menage",
    "prest_fam_petite_enfance": "paje_menage",
    "prest_logement": "aides_logement_menage",
    "prest_precarite_rsa": "rsa_menage",
    "prest_precarite_hand": "aah_menage",
    "prest_precarite_vieil": 'aspa_menage',
    "revdispm": "revenu_disponible",
    "nb_uci": "unites_consommation",
    }


class ErfsFprtoInputComparator(AbstractComparator):
    name = "erfs_fpr"
    period = None
    default_target_variables = [
        "chomage_imposable",
        "loyer",
        # "rag",
        "retraite_imposable",
        # "ric",
        # "rnc",
       "salaire_imposable",
       "statut_occupation_logement",
        ]

    def compute_test_dataframes(self):
        erfs_fpr_survey_collection = SurveyCollection.load(collection = "erfs_fpr")
        # infer names of the survey and data tables
        assert self.period is not None
        year = int(self.period)
        table_by_name = build_table_by_name(year, erfs_fpr_survey_collection)

        log.debug(f"Loading tables for year {year} [{table_by_name.values()}]")

        # load survey and tables
        survey = erfs_fpr_survey_collection.get_survey(table_by_name['survey'])

        fpr_individu = survey.get_values(table = table_by_name['fpr_individu'], ignorecase = True)
        fpr_menage = survey.get_values(table = table_by_name['fpr_menage'], ignorecase = True)

        openfisca_survey_collection = SurveyCollection.load(collection = "openfisca_erfs_fpr")
        openfisca_survey = openfisca_survey_collection.get_survey(f"openfisca_erfs_fpr_{year}")
        openfisca_individu = openfisca_survey.get_values(table = f"individu_{year}")
        openfisca_menage = openfisca_survey.get_values(table = f"menage_{year}")

        input_dataframe_by_entity = {
            "individu": openfisca_individu,
            "menage": openfisca_menage,
            }

        fpr_menage.loyer = 12 * fpr_menage.loyer

        target_dataframe_by_entity = {
            "individu": fpr_individu.rename(columns = openfisca_by_erfs_fpr_variables),
            "menage": fpr_menage.rename(columns = openfisca_by_erfs_fpr_variables),
            }

        return input_dataframe_by_entity, target_dataframe_by_entity


@click.command()
@click.option('-b', '--browse', is_flag = True, help = "Browse results", default = False, show_default = True)
@click.option('-l', '--load', is_flag = True, default = False, help = "Load backup results", show_default = True)
@click.option('-v', '--verbose', is_flag = True, default = False, help = "Increase output verbosity", show_default = True)
@click.option('-d', '--debug', is_flag = True, default = False, help = "Use python debugger", show_default = True)
@click.option('-p', '--period', default = REFERENCE_YEAR, help = "period(s) to treat", show_default = True)
@click.option('-t', '--target-variables', default = None, help = "target variables to inspect (None means all)", show_default = True)
@click.option('-u', '--rebuild', is_flag = True, default = False, help = "Rebuild test data", show_default = True)
@click.option('-s', '--summary', is_flag = True, default = False, help = "Produce summary figuress", show_default = True)
def compare(browse = False, load = False, verbose = True, debug = True, target_variables = None, period = None, rebuild = False, summary = False):
    """Compare openfisca-france-data simulation to erfs-fpr by generating comparison data and graphs.

    Data can be explored using D-Tale and graphs are saved as pdf files.
    """
    comparator = ErfsFprtoInputComparator()
    comparator.period = period
    comparator.compare(browse, load, verbose, debug, target_variables, period, rebuild, summary)



#! /usr/bin/env python
import gc
import numpy as np
import logging


from openfisca_survey_manager.temporary import temporary_store_decorator
from openfisca_france_data.utils import assert_dtype
from openfisca_france_data.erfs.input_data_builder.base import (
    year_specific_by_generic_data_frame_name
    )
from openfisca_survey_manager.survey_collections import SurveyCollection


log = logging.getLogger(__name__)


def create_actrec_variable(indivim):
    """
    Création de la variables actrec
    pour activité recodée comme preconisé par l'INSEE p84 du guide méthodologique de l'ERFS
    """
    assert "actrec" not in indivim.columns
    indivim["actrec"] = np.nan
    # Attention : Pas de 6, la variable recodée de l'INSEE (voit p84 du guide methodo), ici \
    # la même nomenclature à été adopée
    # 3: contrat a durée déterminée
    indivim.loc[indivim.acteu == 1, 'actrec'] = 3
    # 8: femme (homme) au foyer, autre inactif
    indivim.loc[indivim.acteu == 3, 'actrec'] = 8
    # 1: actif occupé non salarié
    filter1 = (indivim.acteu == 1) & (indivim.stc.isin([1, 3]))  # actifs occupés non salariés à son compte ou pour un
    indivim.loc[filter1, 'actrec'] = 1                              # membre de sa famille
    # 2: salarié pour une durée non limitée
    filter2 = (indivim.acteu == 1) & (((indivim.stc == 2) & (indivim.contra == 1)) | (indivim.titc == 2))
    indivim.loc[filter2, 'actrec'] = 2
    # 4: au chomage
    filter4 = (indivim.acteu == 2) | ((indivim.acteu == 3) & (indivim.mrec == 1))
    indivim.loc[filter4, 'actrec'] = 4
    # 5: élève étudiant , stagiaire non rémunéré
    filter5 = (indivim.acteu == 3) & ((indivim.forter == 2) | (indivim.rstg == 1))
    indivim.loc[filter5, 'actrec'] = 5
    # 7: retraité, préretraité, retiré des affaires unchecked
    filter7 = (indivim.acteu == 3) & ((indivim.retrai == 1) | (indivim.retrai == 2))
    indivim.loc[filter7, 'actrec'] = 7
    # 9: probablement enfants de - de 16 ans TODO: check that fact in database and questionnaire
    indivim.loc[indivim.acteu == 0, 'actrec'] = 9

    assert indivim.actrec.notnull().all()
    indivim.actrec = indivim.actrec.astype("int8")
    assert_dtype(indivim.actrec, "int8")

    assert (indivim.actrec != 6).all(), 'actrec ne peut pas être égale à 6'
    assert indivim.actrec.isin(range(1, 10)).all(), 'actrec values are outside the interval [1, 9]'


def create_variable_locataire(menagem):
    # Locataire
    menagem["locataire"] = menagem.so.isin([3, 4, 5])
    assert_dtype(menagem.locataire, "bool")


def check_integer_dtype(indivim, var_list):
    for var in var_list:
        assert np.issubdtype(indivim[var].dtype, np.integer), \
            "Variable {} dtype is {} and should be an integer".format(var, indivim[var].dtype)


def manually_remove_noindiv_errors(indivim):
    '''
    This method is here because some oddities can make it through the controls throughout the procedure
    It is here to remove all these individual errors that compromise the process.
    '''
    if year == 2006:
        indivim.lien[indivim.noindiv == 603018905] = 2
        indivim.noimer[indivim.noindiv == 603018905] = 1
        log.info("{}".format(indivim[indivim.noindiv == 603018905].to_string()))


@temporary_store_decorator(file_name = "erfs")
def merge_tables(temporary_store = None, year = None):
    """
    Création des tables ménages et individus concaténée (merged)
    """
    # Prepare the some useful merged tables

    assert temporary_store is not None
    assert year is not None
    # load data
    erfs_survey_collection = SurveyCollection.load(
        collection = 'erfs', config_files_directory = config_files_directory)

    year_specific_by_generic = year_specific_by_generic_data_frame_name(year)
    survey = erfs_survey_collection.get_survey('erfs_{}'.format(year))
    erfmen = survey.get_values(table = year_specific_by_generic["erf_menage"])
    eecmen = survey.get_values(table = year_specific_by_generic["eec_menage"])
    erfind = survey.get_values(table = year_specific_by_generic["erf_indivi"])
    eecind = survey.get_values(table = year_specific_by_generic["eec_indivi"])

    # travail sur la cohérence entre les bases
    noappar_m = eecmen[~(eecmen.ident.isin(erfmen.ident.values))].copy()

    noappar_i = eecmen[~(eecind.ident.isin(erfind.ident.values))].copy()
    noappar_i = noappar_i.drop_duplicates(subset = 'ident', take_last = True)
    # TODO: vérifier qu'il n'y a théoriquement pas de doublon

    difference = set(noappar_i.ident).symmetric_difference(noappar_m.ident)
    intersection = set(noappar_i.ident) & set(noappar_m.ident)
    log.info("There are {} differences and {} intersections".format(len(difference), len(intersection)))
    del noappar_i, noappar_m, difference, intersection
    gc.collect()

    # fusion enquete emploi et source fiscale
    menagem = erfmen.merge(eecmen)
    indivim = eecind.merge(erfind, on = ['noindiv', 'ident', 'noi'], how = "inner")
    var_list = [
        'acteu',
        'agepr',
        'cohab',
        'contra',
        'encadr',
        'forter',
        'lien',
        'mrec',
        'naia',
        'noicon',
        'noimer',
        'noiper',
        'prosa',
        'retrai',
        'rstg',
        'statut',
        'stc',
        'otitc',
        'txtppb',
        ]
    check_integer_dtype(indivim, var_list)

    create_actrec_variable(indivim)
    create_variable_locataire(menagem)
    menagem = menagem.merge(
        indivim.loc[indivim.lpr == 1, ['ident', 'ddipl']].copy()
        )
    manually_remove_noindiv_errors(indivim)

    temporary_store['menagem_{}'.format(year)] = menagem
    del eecmen, erfmen, menagem
    gc.collect()
    temporary_store['indivim_{}'.format(year)] = indivim
    del erfind, eecind


@temporary_store_decorator(file_name = "erfs")
def create_enfants_a_naitre(temporary_store = None, year = None):
    '''
    '''
    assert temporary_store is not None
    assert year is not None

    erfs_survey_collection = SurveyCollection.load(
        collection = 'erfs', config_files_directory = config_files_directory)
    survey = erfs_survey_collection.get_survey('erfs_{}'.format(year))
    # Enfant à naître (NN pour nouveaux nés)
    individual_vars = [
        'acteu',
        'agepr',
        'cohab',
        'contra',
        'forter',
        'ident',
        'lien',
        'lpr',
        'mrec',
        'naia',
        'naim',
        'noi',
        'noicon',
        'noimer',
        'noindiv',
        'noiper',
        'retrai',
        'rga',
        'rstg',
        'sexe',
        'stc',
        'titc',
        ]
    year_specific_by_generic = year_specific_by_generic_data_frame_name(year)
    eeccmp1 = survey.get_values(table = year_specific_by_generic["eec_cmp_1"], variables = individual_vars)
    eeccmp2 = survey.get_values(table = year_specific_by_generic["eec_cmp_2"], variables = individual_vars)
    eeccmp3 = survey.get_values(table = year_specific_by_generic["eec_cmp_3"], variables = individual_vars)
    tmp = eeccmp1.merge(eeccmp2, how = "outer")
    enfants_a_naitre = tmp.merge(eeccmp3, how = "outer")

    # optimisation des types? Controle de l'existence en passant
    # pourquoi pas des int quand c'est possible
    # TODO: minimal dtype TODO: shoudln't be here
    for var in individual_vars:
        assert_dtype(enfants_a_naitre[var], 'float')
    del eeccmp1, eeccmp2, eeccmp3, individual_vars, tmp
    gc.collect()

    # création de variables
    enfants_a_naitre['declar1'] = ''
    enfants_a_naitre['noidec'] = 0
    enfants_a_naitre['ztsai'] = 0
    enfants_a_naitre['year'] = year
    enfants_a_naitre.year = enfants_a_naitre.year.astype("float32")  # TODO: should be an integer but NaN are present
    enfants_a_naitre['agepf'] = enfants_a_naitre.year - enfants_a_naitre.naia
    enfants_a_naitre.loc[enfants_a_naitre.naim >= 7,'agepf'] -= 1
    enfants_a_naitre['actrec'] = 9
    enfants_a_naitre['quelfic'] = 'ENF_NN'
    enfants_a_naitre['persfip'] = ""

    # TODO: deal with agepf
    for series_name in ['actrec', 'noidec', 'ztsai']:
        assert_dtype(enfants_a_naitre[series_name], "int")

    # selection
    enfants_a_naitre = enfants_a_naitre[
        (
            (enfants_a_naitre.naia == enfants_a_naitre.year) & (enfants_a_naitre.naim >= 10)
            ) | (
                (enfants_a_naitre.naia == enfants_a_naitre.year + 1) & (enfants_a_naitre.naim <= 5)
                )
        ].copy()

    temporary_store["enfants_a_naitre_{}".format(year)] = enfants_a_naitre

if __name__ == '__main__':
    log.info('Entering 01_pre_proc')
    import sys
    import time
    logging.basicConfig(level = logging.INFO, stream = sys.stdout)
    deb = time.clock()
    year = 2009
    merge_tables(year = year)
    create_enfants_a_naitre(year = year)
    log.info("etape 01 pre-processing terminee en {}".format(time.clock() - deb))



#from numpy import where, NaN, random
from openfisca_france.data.erf.build_survey import show_temp, load_temp, save_temp
from openfisca_france.data.erf.build_survey.utils import print_id, control, check_structure
from pandas import read_csv, HDFStore
from openfisca_france import DATA_SOURCES_DIR
import os


def final_check(year=2006):
    test_filename = os.path.join(DATA_SOURCES_DIR, "test.h5")
    survey_filename = os.path.join(DATA_SOURCES_DIR, "survey.h5")

    store = HDFStore(test_filename)
    survey = HDFStore(survey_filename)

    final2 = store.get('survey_2006')
    finalT = survey.get('survey_2006')

    varlist = [
        'adeben',
        'adfdap',
        'amois',
        'ancchom',
        'ancentr',
        'anciatm',
        'ancrech',
        'anref',
        'contra',
        'datant',
        'dimtyp',
        'ident',
        'idfoy'
        'noi',
        'nondic',
        'rabs',
        'RABSP',
        'RAISTP',
        'raistp',
        'rdem',
        'retrai',
        'sitant',
        'sp10',
        'sp11',
        'stc',
        'TXTPPB',
        ]

    for i in range(0, 10):
        varname = 'sp0' + str(i)
        varlist.append(varname)

    varlist = set(varlist)
    columns = final2.columns
    columns = set(columns)

    print(varlist.difference(columns))
    print(final2.loc[
        final2.idfoy == 603018901,
        ['idfoy', 'quifoy', 'idfam', 'quifam', 'idmen', 'quimen', 'noi']
        ].to_string()
        )

    return


if __name__ == '__main__':
    final_check()



from numpy import arange


try:
    from openfisca_survey_manager.statshelpers import weighted_quantiles
except ImportError:
    weighted_quantiles = None

try:
    from openfisca_survey_manager.statshelpers import mark_weighted_percentiles
except ImportError:
    mark_weighted_percentiles = None

from openfisca_france_data.model.base import *  # noqa analysis:ignore


class assiette_csg_salaire(Variable):
    value_type = float
    entity = Individu
    label = "Assiette CSG salaires"
    definition_period = MONTH

    def formula(individu, period, parameters):
        assiette_csg_abattue = individu('assiette_csg_abattue', period)
        assiette_csg_non_abattue = individu('assiette_csg_non_abattue', period)
        plafond_securite_sociale = individu('plafond_securite_sociale', period)
        abattement = parameters(period.start).prelevements_sociaux.contributions_sociales.csg.activite.deductible.abattement
        assiette = assiette_csg_abattue - abattement.calc(
            assiette_csg_abattue,
            factor = plafond_securite_sociale,
            round_base_decimals = 2,
            ) + assiette_csg_non_abattue
        return assiette


class assiette_csg_retraite(Variable):
    value_type = float
    entity = Individu
    label = "Assiette CSG retraite"
    definition_period = MONTH

    def formula(individu, period, parameters):
        retraite_brute = individu('retraite_brute', period)
        taux_csg_remplacement = individu('taux_csg_remplacement', period)
        return retraite_brute * (taux_csg_remplacement >= 2)


class assiette_csg_chomage(Variable):
    value_type = float
    entity = Individu
    label = "Assiette CSG chomage"
    definition_period = MONTH

    def formula(individu, period):
        chomage_brut = individu('chomage_brut', period)
        taux_csg_remplacement = individu('taux_csg_remplacement', period)
        return chomage_brut * (taux_csg_remplacement >= 2)


class csg_salaire(Variable):
    value_type = float
    entity = Individu
    label = "CSG salaire"
    definition_period = MONTH

    def formula(individu, period):
        return  (
            individu('csg_deductible_salaire', period)
            + individu('csg_imposable_salaire', period)
            )


class csg_non_salarie(Variable):
    value_type = float
    entity = Individu
    label = "CSG non salarié"
    definition_period = YEAR

    def formula(individu, period):
        return  (
            individu('csg_deductible_non_salarie', period)
            + individu('csg_imposable_non_salarie', period)
            )


class csg_remplacement(Variable):
    value_type = float
    entity = Individu
    label = "CSG sur les revenus de remplacement"
    definition_period = MONTH

    def formula(individu, period):
        return  (
            individu('csg_imposable_retraite', period)
            + individu('csg_deductible_retraite', period)
            + individu('csg_imposable_chomage', period)
            + individu('csg_deductible_chomage', period)
            )


class impot_revenu(Variable):
    value_type = float
    entity = FoyerFiscal
    label = "Impôt sur le revenu"
    definition_period = YEAR

    def formula(foyer_fiscal, period):
        return  (
            foyer_fiscal('irpp_economique', period)
            + foyer_fiscal('prelevement_forfaitaire_liberatoire', period)
            + foyer_fiscal('prelevement_forfaitaire_unique_ir', period)
            + foyer_fiscal('ir_pv_immo', period)
            )


class decile(Variable):
    value_type = Enum
    possible_values = Deciles
    default_value = Deciles.hors_champs
    entity = Menage
    label = "Décile de niveau de vie disponible"
    definition_period = YEAR

    def formula(menage, period):
        menage_ordinaire = menage('menage_ordinaire', period)
        niveau_de_vie = menage('niveau_de_vie', period)
        wprm = menage('wprm', period)
        labels = arange(1, 11)
        method = 2
        if len(wprm) == 1:
            return wprm * 0
        decile, values = mark_weighted_percentiles(
            niveau_de_vie, labels, wprm * menage_ordinaire, method, return_quantiles = True)
        del values
        return decile * menage_ordinaire


class centile(Variable):
    value_type = Enum
    possible_values = Deciles
    default_value = Deciles.hors_champs
    entity = Menage
    label = "Centile de niveau de vie disponible"
    definition_period = YEAR

    def formula(menage, period):
        menage_ordinaire = menage('menage_ordinaire', period)
        niveau_de_vie = menage('niveau_de_vie', period)
        wprm = menage('wprm', period)
        labels = arange(1, 101)
        method = 2
        if len(wprm) == 1:
            return wprm * 0
        centile, values = mark_weighted_percentiles(
            niveau_de_vie, labels, wprm * menage_ordinaire, method, return_quantiles = True)
        del values
        return centile * menage_ordinaire


class decile_rfr(Variable):
    value_type = Enum
    possible_values = Deciles
    default_value = Deciles.hors_champs
    entity = FoyerFiscal
    label = "Décile de revenu fiscal de référence"
    definition_period = YEAR

    def formula(foyer_fiscal, period):
        rfr = foyer_fiscal('rfr', period)
        weight_foyers = foyer_fiscal('weight_foyers', period)
        menage_ordinaire_foyers_fiscaux = foyer_fiscal('menage_ordinaire_foyers_fiscaux', period)
        labels = arange(1, 11)
        # Alternative method
        # method = 2
        # decile, values = mark_weighted_percentiles(niveau_de_vie, labels, pondmen, method, return_quantiles = True)
        decile, values = weighted_quantiles(rfr, labels, weight_foyers * menage_ordinaire_foyers_fiscaux, return_quantiles = True)
        return decile


class centile_rfr(Variable):
    value_type = Enum
    possible_values = Deciles
    default_value = Deciles.hors_champs
    entity = FoyerFiscal
    label = "Centile de revenu fiscal de référence"
    definition_period = YEAR

    def formula(foyer_fiscal, period):
        rfr = foyer_fiscal('rfr', period)
        weight_foyers = foyer_fiscal('weight_foyers', period)
        menage_ordinaire_foyers_fiscaux = foyer_fiscal('menage_ordinaire_foyers_fiscaux', period)
        labels = arange(1, 101)
        centile, values = weighted_quantiles(rfr, labels, weight_foyers * menage_ordinaire_foyers_fiscaux, return_quantiles = True)
        return centile


class decile_rfr_par_part(Variable):
    value_type = Enum
    possible_values = Deciles
    default_value = Deciles.hors_champs
    entity = FoyerFiscal
    label = "Décile de revenu fiscal de référence par part fiscale"
    definition_period = YEAR

    def formula(foyer_fiscal, period):
        rfr = foyer_fiscal('rfr', period)
        nbptr = foyer_fiscal('nbptr', period)
        weight_foyers = foyer_fiscal('weight_foyers', period)
        menage_ordinaire_foyers_fiscaux = foyer_fiscal('menage_ordinaire_foyers_fiscaux', period)
        labels = arange(1, 11)
        # Alternative method
        # method = 2
        # decile, values = mark_weighted_percentiles(niveau_de_vie, labels, pondmen, method, return_quantiles = True)
        decile, values = weighted_quantiles(
            rfr / nbptr, labels, weight_foyers * menage_ordinaire_foyers_fiscaux, return_quantiles = True)
        return decile


class centile_rfr_par_part(Variable):
    value_type = Enum
    possible_values = Deciles
    default_value = Deciles.hors_champs
    entity = FoyerFiscal
    label = "Centile de revenu fiscal de référence par part fiscale"
    definition_period = YEAR

    def formula(foyer_fiscal, period):
        rfr = foyer_fiscal('rfr', period)
        nbptr = foyer_fiscal('nbptr', period)
        weight_foyers = foyer_fiscal('weight_foyers', period)
        menage_ordinaire_foyers_fiscaux = foyer_fiscal('menage_ordinaire_foyers_fiscaux', period)
        labels = arange(1, 101)
        centile, values = weighted_quantiles(
            rfr / nbptr, labels, weight_foyers * menage_ordinaire_foyers_fiscaux, return_quantiles = True)
        return centile


class pauvre40(Variable):
    value_type = bool
    entity = Menage
    label = "Ménage en dessous du seuil de pauvreté à 40%"
    definition_period = YEAR

    def formula(menage, period):
        menage_ordinaire = menage('menage_ordinaire', period)
        nivvie = menage('nivvie', period)
        wprm = menage('wprm', period)
        labels = arange(1, 3)
        method = 2
        if len(wprm) == 1:
            return wprm * 0
        percentile, values = mark_weighted_percentiles(nivvie, labels, wprm * menage_ordinaire, method, return_quantiles = True)
        threshold = .4 * values[1]
        return (nivvie <= threshold) * menage_ordinaire


class pauvre50(Variable):
    value_type = bool
    entity = Menage
    label = "Ménage en dessous du seuil de pauvreté à 50%"
    definition_period = YEAR


    def formula(menage, period):
        menage_ordinaire = menage('menage_ordinaire', period)
        nivvie = menage('nivvie', period)
        wprm = menage('wprm', period)
        labels = arange(1, 3)
        method = 2
        if len(wprm) == 1:
            return wprm * 0

        percentile, values = mark_weighted_percentiles(nivvie, labels, wprm * menage_ordinaire, method, return_quantiles = True)
        threshold = .5 * values[1]
        return (nivvie <= threshold) * menage_ordinaire


class pauvre60(Variable):
    value_type = bool
    entity = Menage
    label = "Ménage en dessous du seuil de pauvreté à 60%"
    definition_period = YEAR

    def formula(menage, period):
        menage_ordinaire = menage('menage_ordinaire', period)
        nivvie = menage('nivvie', period)
        wprm = menage('wprm', period)
        labels = arange(1, 3)
        method = 2
        if len(wprm) == 1:
            return wprm * 0

        percentile, values = mark_weighted_percentiles(nivvie, labels, wprm * menage_ordinaire, method, return_quantiles = True)
        threshold = .6 * values[1]
        return (nivvie <= threshold) * menage_ordinaire


class salaire_brut(Variable):
    value_type = int
    entity = Individu
    label = u"Salaire brut (y compris fonction publique)"
    set_input = set_input_divide_by_period
    definition_period = MONTH

    def formula(individu, period):
        salaire_de_base = individu('salaire_de_base', period)
        primes_salaires = individu('primes_salaires', period)
        primes_fonction_publique = individu('primes_fonction_publique', period)
        indemnite_residence = individu('indemnite_residence', period)
        supplement_familial_traitement = individu('supplement_familial_traitement', period)
        remuneration_principale = individu('remuneration_principale', period)
        indemnite_fin_contrat = individu('indemnite_fin_contrat', period)
        complementaire_sante_salarie = individu('complementaire_sante_salarie', period)
        indemnite_compensatrice_csg = individu('indemnite_compensatrice_csg', period)

        return (
            salaire_de_base
            + primes_salaires
            + remuneration_principale
            + primes_fonction_publique
            + indemnite_residence
            + supplement_familial_traitement
            + indemnite_fin_contrat
            + complementaire_sante_salarie
            + indemnite_compensatrice_csg
            )



#! /usr/bin/env python
import gc

import logging
import numpy as np
from numpy import where, NaN, random
from pandas import read_csv
import os



from openfisca_survey_manager.temporary import temporary_store_decorator
from openfisca_france_data.utils import (
    check_structure,
    control,
    normalizes_roles_in_entity,
    id_formatter,
    print_id,
    rectify_dtype,
    set_variables_default_value,
    )

log = logging.getLogger(__name__)


@temporary_store_decorator(file_name = 'erfs')
def final(temporary_store = None, year = None, check = True):

    assert temporary_store is not None
    assert year is not None

    log.info(u'08_final: derniers réglages')

    # On définit comme célibataires les individus dont on n'a pas retrouvé la déclaration
    final = temporary_store['final_{}'.format(year)]

    if year == 2009:
        final = normalizes_roles_in_entity(final, 'foy')
        final = normalizes_roles_in_entity(final, 'men')

    if check:
        check_structure(final)

    final.statmarit.fillna(2, inplace = True)

    # activite des fip
    log.info('    gestion des FIP de final')
    final_fip = final.loc[final.quelfic.isin(["FIP", "FIP_IMP"]), ["choi", "sali", "alr", "rsti", "age"]].copy()

    log.info(set(["choi", "sali", "alr", "rsti"]).difference(set(final_fip.columns)))
    for var in ["choi", "sali", "alr", "rsti"]:
        final_fip[var].fillna(0, inplace = True)
        assert final_fip[var].notnull().all(), "Some NaN are remaining in column {}".format(var)

    final_fip["activite"] = 2  # TODO comment choisr la valeur par défaut ?
    final_fip.activite = where(final_fip.choi > 0, 1, final_fip.activite)
    final_fip.activite = where(final_fip.sali > 0, 0, final_fip.activite)
    final_fip.activite = where(final_fip.age > 21, 2, final_fip.activite)  # ne peuvent être rattachés que les étudiants

    final.update(final_fip)
    temporary_store['final_{}'.format(year)] = final
    log.info("final has been updated with fip")

    menagem = temporary_store['menagem_{}'.format(year)]
    assert 'ident' in menagem.columns
    assert 'so' in menagem.columns
    menagem.rename(
        columns = dict(ident = "idmen", so = "statut_occupation"),
        inplace = True
        )
    menagem["cstotpragr"] = np.floor(menagem["cstotpr"] / 10)

    # 2008 tau99 removed TODO: check ! and check incidence
    variables = [
        "champm",
        "cstotpragr",
        "ddipl",
        "idmen",
        "nbinde",
        "pol99",
        "reg",
        "statut_occupation",
        "tau99",
        "tu99",
        "typmen15",
        "wprm",
        "zthabm",
        ]
    if year == 2008:
        variables.remove("tau99")
    famille_variables = ["m_afeamam", "m_agedm", "m_clcam", "m_colcam", 'm_mgamm', 'm_mgdomm']

    if 'loyer' in menagem.columns:
        variables.append('loyer')
# if ("naf16pr" %in% names(menagem)) {
#   naf16pr <- factor(menagem$naf16pr)
#   levels(naf16pr) <-  0:16
#   menagem$naf16pr <- as.character(naf16pr)
#   menagem[is.na(menagem$naf16pr), "naf16pr" ] <- "-1"  # Sans objet
#   variables <- c(variables,"naf16pr")
# } else if ("nafg17npr" %in% names(menagem)) {
#   # TODO: pb in 2008 with xx
#   if (year == "2008"){
#     menagem[ menagem$nafg17npr == "xx" & !is.na(menagem$nafg17npr), "nafg17npr"] <- "00"
#   }
#   nafg17npr <- factor(menagem$nafg17npr)
#   levels(nafg17npr) <-  0:17
#   menagem$nafg17npr <- as.character(nafg17npr)
#   menagem[is.na(menagem$nafg17npr), "nafg17npr" ] <- "-1"  # Sans objet
# }
#
# TODO: pytohn translation needed
#    if "naf16pr" in menagem.columns:
#        naf16pr <- factor(menagem$naf16pr)
#   levels(naf16pr) <-  0:16
#   menagem$naf16pr <- as.character(naf16pr)
#   menagem[is.na(menagem$naf16pr), "naf16pr" ] <- "-1"  # Sans objet
#   variables <- c(variables,"naf16pr")
# } else if ("nafg17npr" %in% names(menagem)) {
#   # TODO: pb in 2008 with xx
#   if (year == "2008"){
#     menagem[ menagem$nafg17npr == "xx" & !is.na(menagem$nafg17npr), "nafg17npr"] <- "00"
#   }
#   nafg17npr <- factor(menagem$nafg17npr)
#   levels(nafg17npr) <-  0:17
#   menagem$nafg17npr <- as.character(nafg17npr)
#   menagem[is.na(menagem$nafg17npr), "nafg17npr" ] <- "-1"  # Sans objet
# }
    # TODO: 2008tau99 is not present should be provided by 02_loy.... is it really needed
    all_variables = variables + famille_variables

    log.info("liste de toutes les variables : {}".format(all_variables))
    log.info(menagem.info())
    available_variables = list(set(all_variables).intersection(set(menagem.columns)))
    log.info("liste des variables à extraire de menagem: {}".format(available_variables))
    loyersMenages = menagem[available_variables].copy()
#
# # Recodage de typmen15: modalités de 1:15
# table(loyersMenages$typmen15, useNA="ifany")
# loyersMenages <- within(loyersMenages, {
#   typmen15[typmen15==10 ] <- 1
#   typmen15[typmen15==11 ] <- 2
#   typmen15[typmen15==21 ] <- 3
#   typmen15[typmen15==22 ] <- 4
#   typmen15[typmen15==23 ] <- 5
#   typmen15[typmen15==31 ] <- 6
#   typmen15[typmen15==32 ] <- 7
#   typmen15[typmen15==33 ] <- 8
#   typmen15[typmen15==41 ] <- 9
#   typmen15[typmen15==42 ] <- 10
#   typmen15[typmen15==43 ] <- 11
#   typmen15[typmen15==44 ] <- 12
#   typmen15[typmen15==51 ] <- 13
#   typmen15[typmen15==52 ] <- 14
#   typmen15[typmen15==53 ] <- 15
# })
#
#
# TODO: MBJ UNNECESSARY ?
#
# # Pb avec ddipl, pas de modalités 2: on décale les chaps >=3
# # Cependant on fait cela après avoir fait les traitement suivants
# table(loyersMenages$ddipl, useNA="ifany")
# # On convertit les ddipl en numeric
# loyersMenages$ddipl <- as.numeric(loyersMenages$ddipl)
# table(loyersMenages$ddipl, useNA="ifany")
# #   On met les non renseignés ie, NA et "" à sans diplome (modalité 7)
# loyersMenages[is.na(loyersMenages$ddipl), "ddipl"] <- 7
#
# loyersMenages[loyersMenages$ddipl>1, "ddipl"] <- loyersMenages$ddipl[loyersMenages$ddipl>1]-1
#
    log.info("loyersMenages \n {}".format(loyersMenages.info()))
    loyersMenages.ddipl = where(loyersMenages.ddipl.isnull(), 7, loyersMenages.ddipl)
    loyersMenages.ddipl = where(loyersMenages.ddipl > 1,
                                loyersMenages.ddipl - 1,
                                loyersMenages.ddipl)
    loyersMenages.ddipl = loyersMenages.ddipl.astype("int32")

    final['act5'] = NaN
    final.act5 = where(final.actrec == 1, 2, final.act5)  # indépendants
    final.act5 = where(final.actrec.isin([2, 3]), 1, final.act5)  # salariés

    final.act5 = where(final.actrec == 4, 3, final.act5)  # chômeur
    final.act5 = where(final.actrec == 7, 4, final.act5)  # retraité
    final.act5 = where(final.actrec == 8, 5, final.act5)  # autres inactifs

    log.info("Valeurs prises par act5: \n {}".format(final.act5.value_counts()))

#     assert final.act5.notnull().all(), 'there are NaN inside final.act5'
# final$wprm <- NULL # with the intention to extract wprm from menage to deal with FIPs
# final$taxe_habitation <- final$zthabm # rename zthabm to taxe_habitation
# final$zthabm <- NULL
#
# final2 <- merge(final, loyersMenages, by="idmen", all.x=TRUE)
    log.info('    création de final2')
    del final["wprm"]
    gc.collect()
    final.rename(columns = dict(zthabm = "taxe_habitation"), inplace = True)  # rename zthabm to taxe_habitation
    final2 = final.merge(loyersMenages, on = "idmen", how = "left")  # TODO: Check
    log.info("{}".format(loyersMenages.head()))
    gc.collect()
    print_id(final2)
# # TODO: merging with patrimoine
    log.info('    traitement des zones apl')
    import pkg_resources
    openfisca_france_data_location = pkg_resources.get_distribution('openfisca-france-data').location
    zone_apl_imputation_data_file_path = os.path.join(
        openfisca_france_data_location,
        'openfisca_france_data',
        'zone_apl_data',
        'zone_apl',
        'zone_apl_imputation_data.csv',
        )
    apl_imp = read_csv(zone_apl_imputation_data_file_path)

    log.info("{}".format(apl_imp.head(10)))
    if year == 2008:
        zone_apl = final2.xs(["tu99", "pol99", "reg"], axis = 1)
    else:
        zone_apl = final2.xs(["tu99", "pol99", "tau99", "reg"], axis = 1)

    for i in range(len(apl_imp["TU99"])):
        tu = apl_imp["TU99"][i]
        pol = apl_imp["POL99"][i]
        tau = apl_imp["TAU99"][i]
        reg = apl_imp["REG"][i]

    if year == 2008:
        indices = (final2["tu99"] == tu) & (final2["pol99"] == pol) & (final2["reg"] == reg)
        selection = (apl_imp["TU99"] == tu) & (apl_imp["POL99"] == pol) & (apl_imp["REG"] == reg)
    else:
        indices = (final2["tu99"] == tu) & (final2["pol99"] == pol) & (final2["tau99"] == tau) & (final2["reg"] == reg)
        selection = (
            (apl_imp["TU99"] == tu) &
            (apl_imp["POL99"] == pol) &
            (apl_imp["TAU99"] == tau) &
            (apl_imp["REG"] == reg)
            )

    z = random.uniform(size=indices.sum())
    log.info(len(z))
    log.info(len(indices))
    log.info(len(indices) / len(z))
    probs = apl_imp[["proba_zone1", "proba_zone2"]][selection].copy()
    log.info(probs)
    log.info(probs['proba_zone1'].values)
    proba_zone_1 = probs['proba_zone1'].values[0]
    proba_zone_2 = probs['proba_zone2'].values[0]

    final2["zone_apl"] = 3
    final2.zone_apl.loc[indices] = (
        1 + (z > proba_zone_1) + (z > (proba_zone_1 + proba_zone_2))
        )
    del indices, probs

    log.info('    performing cleaning on final2')
    log.info('{} sali nuls'.format(len(final2[final2['sali'].isnull()])))
    log.info("{} individus d'âges nuls".format(len(final2[final2.age.isnull()])))
    log.info("longueur de final2 avant purge : {}".format(len(final2)))
    log.info('check doublons : {}'.format(len(final2[final2.duplicated(['noindiv'])])))
    log.info("{}".format(final2.age.isnull().sum()))

    print_id(final2)

    control(final2, debug = True)
    temporary_store['final2'] = final2
    log.info("Nombre de personne d'âge NaN: {} ".format(final2.age.isnull().sum()))
    final2 = final2.drop_duplicates(subset = 'noindiv')

    log.info('    Filter to manage the new 3-tables structures:')
    # On récupère les foyer, famille, ménages qui ont un chef :
    # On ne conserve dans final2 que ces foyers là :
    log.info('final2 avant le filtrage {}'.format(len(final2)))
    print_id(final2)


    liste_fam = final2.loc[final2['quifam'] == 0, 'idfam'].unique()
    log.info("Dropping {} famille".format((~final2.idfam.isin(liste_fam)).sum()))
    final2 = final2.loc[final2.idfam.isin(liste_fam)].copy()

    if check:
        check_structure(final2)

    temporary_store['final2bis'] = final2

    liste_foy = final2.loc[final2['quifoy'] == 0, 'idfoy'].unique()
    log.info("Dropping {} foyers".format((~final2.idfoy.isin(liste_foy)).sum()))
    final2 = final2.loc[final2.idfoy.isin(liste_foy)].copy()


    liste_men = final2.loc[final2['quimen'] == 0, 'idmen'].unique()
    log.info(u"Dropping {} ménages".format((~final2.idmen.isin(liste_men)).sum()))
    final2 = final2.loc[final2.idmen.isin(liste_men)].copy()

    log.info('final2 après le filtrage {}'.format(len(final2)))
    print_id(final2)
    rectify_dtype(final2, verbose = False)
    #    home = os.path.expanduser("~")
    #    test_filename = os.path.join(home, filename + ".h5")
    #    if os.path.exists(test_filename):
    #        import warnings
    #        import datetime
    #        time_stamp = datetime.datetime.now().strftime('%Y_%m_%d_%H_%M')
    #        renamed_file = os.path.join(DATA_SOURCES_DIR, filename + "_" + time_stamp + ".h5")
    #        warnings.warn("A file with the same name already exists \n Renaming current output and saving to " + renamed_file)
    #        test_filename = renamed_file
    data_frame = final2

    temporary_store['final3'] = data_frame

    assert not data_frame.duplicated(['idmen', 'quimen']).any(), 'bad ménages indexing'


    if year == 2006:  # Hack crade pur régler un problème rémanent
        data_frame = data_frame[data_frame.idfam != 602177906].copy()

    for id_variable in ['idfam', 'idfoy', 'idmen', 'noi', 'quifam', 'quifoy', 'quimen']:
        data_frame[id_variable] = data_frame[id_variable].astype('int')

    gc.collect()

    if check:
        check_structure(data_frame)
    temporary_store['final4'] = data_frame

    assert not data_frame.duplicated(['idmen', 'quimen']).any(), 'bad ménages indexing'

    gc.collect()
    for entity_id in ['idmen', 'idfoy', 'idfam']:
        log.info('Reformat ids: {}'.format(entity_id))
        data_frame = id_formatter(data_frame, entity_id)

    assert not data_frame.duplicated(['idmen', 'quimen']).any(), 'bad ménages indexing'

    log.info('Dealing with loyer')
    if 'loyer' in data_frame.columns:
        assert 'loyer' in data_frame.columns
        data_frame.loyer = data_frame.loyer * 12.0
    log.info('Set variables to their default values')
    set_variables_default_value(data_frame, year)
    print_id(data_frame)
    temporary_store['input_{}'.format(year)] = data_frame
    return data_frame


if __name__ == '__main__':
    logging.basicConfig(level = logging.INFO, filename = 'step_08.log', filemode = 'w')
    year = 2009
    final(year = year)



# TODO: THIS NEEDS CLEANING

# this directory contains:
#   - zone_apl_2006.csv : PSDC99, Pop_mun_2006, Zone, REG, POL99, TU99, TAU99
#   - zone_apl.csv : example of row : PARIS    Paris    75012    75056    1
#   - zone_apl_imputation_data : ,TU99,TAU99,REG,POL99,proba_zone1,proba_zone2,proba_zone3


# zone_apl_imputation_data_reader.py uses zone_apl_2006 to build zone_apl_imputation_data

# code_apl in france.data is used by a widget codeAplReader was used to produce it.
# Should be stored here



from openfisca_france_data.erfs_fpr.get_survey_scenario import get_survey_scenario
from openfisca_france_data.reforms.inversion_directe_salaires import inversion_directe_salaires

survey_scenario = get_survey_scenario(year = 2014, reform = inversion_directe_salaires)

data_frame_by_entity = survey_scenario.create_data_frame_by_entity(
    variables = [
        'age',
        'aides_logement',
        'aide_logement_montant_brut',
        'apl',
        'alf',
        'als',
        'statut_occupation_logement',
        'weight_familles',
        'weight_individus',
        'wprm',
        ],
    )

famille = data_frame_by_entity['famille']
individu = data_frame_by_entity['individu']
menage = data_frame_by_entity['menage']

# statut_occupation
statut_occupation_logement_pct = menage.groupby('statut_occupation_logement')['wprm'].sum() / menage.wprm.sum()
print(statut_occupation_logement_pct)
# 2 proprietaire
assert .39 < statut_occupation_logement_pct[2] < .41
# accedant
assert .17 < statut_occupation_logement_pct[1] < .18
# locataire (indiferrencie)
assert .39 < statut_occupation_logement_pct[3:7].sum() < .43

11e10 < (famille.aide_logement_montant_brut * famille.weight_familles).sum()
11e10 < (
    (famille.apl * famille.weight_familles).sum()
    + (famille.als * famille.weight_familles).sum()
    + (famille.alf * famille.weight_familles).sum()
    )



import datetime
import dtale
try:
    import ipdb as pdb
except ImportError:
    import pdb
import logging
from math import ceil
import numpy as np
from pathlib import Path, PurePath
import pandas as pd
import pypandoc
import seaborn as sns
import sys


from openfisca_core import periods

from openfisca_france_data.config import config
from openfisca_france_data.erfs_fpr.get_survey_scenario import get_survey_scenario


log = logging.getLogger(__name__)


def get_entity_original_id(survey_scenario, variable):
    entity = survey_scenario.tax_benefit_system.variables[variable].entity.key
    return "noindiv" if entity == "individu" else "idmen_original"


def compute_result(variable, survey_scenario, target_dataframe):
    result = None
    stats = None
    entity = survey_scenario.tax_benefit_system.variables[variable].entity.key
    entity_original_id = get_entity_original_id(survey_scenario, variable)
    output_variables = [entity_original_id, variable]
    entity_dataframe = survey_scenario.create_data_frame_by_entity(
        variables = output_variables,
        )[entity]
    target = target_dataframe[output_variables].rename(columns = {variable: f"target_{variable}"})

    if f"target_{variable}" not in target:
        return None, None

    result = entity_dataframe.merge(
        target,
        on = entity_original_id,
        how = "outer",
        )

    result[f"diff_{variable}"] = result[variable] - result[f"target_{variable}"]
    result_variables = [entity_original_id, variable, f"diff_{variable}", f"target_{variable}"]
    stats = compute_error_stats(result, variable, entity_original_id = entity_original_id)
    return result, stats


def compute_confidence_interval(data, variable, width = .9, entity_original_id = None):
    """
    Compute confidence interval

    Args:
        data (pandas.DataFrame): The data
        variable (str, optional): The variable name. Defaults to None.
        width (float, optional): Witdh of the symmetruc confidence interval. Defaults to .9.

    Returns:
        [type]: [description]
    """
    assert entity_original_id is not None
    df = pd.DataFrame({
        "signed_values": data[variable].values,
        "noind": data[entity_original_id].values
        })
    df["abs_values"] = df.signed_values.abs()
    in_range_obs = ceil(width * len(df))
    sorted_df = df.sort_values("abs_values")
    in_range_values = sorted_df[:in_range_obs]["signed_values"]
    left = in_range_values.min()
    right = in_range_values.max()
    largest_errors = sorted_df[in_range_obs:].copy().sort_values("abs_values", ascending = False)[["signed_values", "abs_values", "noind"]].copy()

    return left, right, largest_errors


def compute_error_stats(data, variable, entity_original_id):
    assert entity_original_id is not None
    numerical = (
        isinstance(data[variable].values.flat[0], np.integer)
        or isinstance(data[variable].values.flat[0], np.floating)
        )
    if not numerical:
        return

    df = data.loc[
        (data[variable].values != 0.0) | (data[f"target_{variable}"].values != 0.0),
        [variable, f"target_{variable}", entity_original_id]
        ].copy()
    df["relative_error"] = (df[variable] - df[f"target_{variable}"]) / (df[f"target_{variable}"] + (df[f"target_{variable}"] == 0.0) * df[variable])
    if df.empty:
        return

    left, right, largest_errors = compute_confidence_interval(df, "relative_error", entity_original_id = entity_original_id)
    less_than_5_pc_error = (df["relative_error"].abs() <= .05).sum() / len(df)
    less_than_20_pc_error = (df["relative_error"].abs() <= .2).sum() / len(df)
    more_than_80_pc_error = (df["relative_error"].abs() >= .8).sum() / len(df)
    return pd.DataFrame.from_dict({
        "variable": [variable],
        "size": [len(data)],
        "share": [len(df) / len(data)],
        "< 5%": [less_than_5_pc_error],
        "< 20%": [less_than_20_pc_error],
        "> 80%": [more_than_80_pc_error],
        "CI (10%)": [[round(left, 2), round(right, 2)]],
        }).round(2)


def create_output_files(markdown_sections, figures_directory, filename):
    header = r"""---
header-includes:
- \usepackage{caption}
- \usepackage{subcaption}
---
"""
    markdown_sections_joined = header + "".join(markdown_sections)
    with open(PurePath.joinpath(figures_directory, "variables.md"), "w", encoding = "utf-8") as markdown_file:
        markdown_file.write(markdown_sections_joined)

    pypandoc.convert_file(
        str(PurePath.joinpath(figures_directory, "variables.md")),
        "pdf",
        format = "markdown",
        outputfile = str(PurePath.joinpath(figures_directory, f"{filename}.pdf")),
        extra_args = ['--pdf-engine=pdflatex'],
        )


def create_variable_distribution_figures(variable, result, bins = None, figures_directory = None, entity_original_id = None):
    assert entity_original_id is not None
    log.debug(f"create_variable_distribution_figures: Examining {variable}")
    assert figures_directory is not None
    if bins is None:
        bins = 100

    non_both_zeroes = (result[f"{variable}"].fillna(0) != 0) | (result[f"target_{variable}"].fillna(0) != 0)
    non_both_zeroes_count = sum(non_both_zeroes)
    both_zeroes_count = len(result) - non_both_zeroes_count

    melted = result.loc[
        non_both_zeroes,
        [entity_original_id, variable, f"target_{variable}"]
        ].melt(
            id_vars = [entity_original_id],
            value_vars = [f"{variable}", f"target_{variable}"]
            )

    if melted.empty:
        print(f"Cannot create variable_distribution_figures for variable = {variable} because all null values")
        return False

    unique_values_count = melted["value"].nunique()

    bins = unique_values_count if unique_values_count < bins else bins
    print(f"create_variable_distribution_figures (total): variable = {variable}, bins = {bins}")

    melted["value"] = melted["value"].clip(1, melted["value"].max())

    log_scale = bins > 10

    sns_plot = sns.histplot(
        data = melted,
        # palette = "crest",
        alpha = .5,
        bins = bins,
        common_bins = True,
        common_norm = False,
        fill = True,
        hue = "variable",
        linewidth = 0,
        x = "value",
        log_scale = log_scale,
        )

    sns_plot.annotate(
        f"Dropping {both_zeroes_count} null observations ({round(100 * both_zeroes_count / (both_zeroes_count + non_both_zeroes_count))}  %)",
        xy = (0, 1),
        xycoords = "axes fraction",
        xytext = (5, -5),
        textcoords = "offset points",
        ha = "left",
        va = "top"
        )
    filename = f"{variable}.pdf"
    sns_plot.figure.savefig(PurePath.joinpath(figures_directory, filename))
    sns_plot.figure.clf()

    return True


def create_variable_markdown_section(variable, stats, figures_directory):
    if stats is None:
        return None

    variable_pdf_path = PurePath.joinpath(figures_directory, f"{variable}.pdf")

    markdown_section = f"""
## Variable `{variable}`


### Valeurs

![]({variable_pdf_path})

### En niveau

![]({PurePath.joinpath(figures_directory, f"diff_{variable}.pdf")})


""" + stats.drop("variable", axis = 1).to_markdown(index = False)

    table_ecarts_markdown_path = PurePath.joinpath(figures_directory, f"table_ecarts_{variable}.md")
    stats.drop("variable", axis = 1).to_markdown(table_ecarts_markdown_path, index = False)

    pypandoc.convert_file(
        str(table_ecarts_markdown_path),
        "pdf",
        format = "markdown",
        outputfile = str(PurePath.joinpath(figures_directory, f"table_ecarts_{variable}.pdf")),
        extra_args = ['--pdf-engine=pdflatex'],
        )

    return markdown_section


def create_variable_markdown_summary_section(variable, stats, figures_directory):
    if stats is None:
        return None
    markdown_section = f"""

## Variable `{variable}`

![]({PurePath.joinpath(figures_directory, f"{variable}.pdf")}){{width=50% height=35%}}
![]({PurePath.joinpath(figures_directory, f"diff_{variable}.pdf")}){{width=50% height=35%}}
\\begin{{figure}}[!h]
\\begin{{subfigure}}[t]{{0.5\\textwidth}}
\\caption{{Distributions des valeurs}}
\\end{{subfigure}}
\\hfill
\\begin{{subfigure}}[t]{{0.5\\textwidth}}
\\caption{{Distributions des écarts}}
\\end{{subfigure}}
\\end{{figure}}

"""
    return markdown_section


def create_diff_variable_distribution_figures(variable, result, bins = None, figures_directory = None, entity_original_id = None):
    assert entity_original_id is not None
    numerical = (
        isinstance(result[f"{variable}"].values.flat[0], np.integer)
        or isinstance(result[f"{variable}"].values.flat[0], np.floating)
        )
    if not numerical:
        return

    assert figures_directory is not None
    if bins is None:
        bins = 100

    non_both_zeroes = (result[f"{variable}"].fillna(0) != 0) | (result[f"target_{variable}"].fillna(0) != 0)
    non_both_zeroes_count = sum(non_both_zeroes)
    both_zeroes_count = len(result) - non_both_zeroes_count

    unique_values_count = result[f"diff_{variable}"].nunique()
    bins == unique_values_count if unique_values_count < bins else bins

    data = result.loc[non_both_zeroes]

    if data.empty:
        print(f"Cannot create diff variable_distribution_figures for variable = {variable} because all null values")
        return

    print(f"create_diff_variable_distribution_figures (total): variable = {variable}, bins = {bins}")
    sns_plot = sns.histplot(data, x = f"diff_{variable}", stat = "probability", bins = bins)

    sns_plot.annotate(
        f"Dropping {both_zeroes_count} null observations ({round(100 * both_zeroes_count / (both_zeroes_count + non_both_zeroes_count))}  %)",
        xy = (0, 1),
        xycoords = "axes fraction",
        xytext = (5, -5),
        textcoords = "offset points",
        ha = "left",
        va = "top"
        )
    filename = f"diff_{variable}.pdf"
    sns_plot.figure.savefig(PurePath.joinpath(figures_directory, filename))
    sns_plot.figure.clf()
    return


class AbstractComparator(object):
    name = None
    default_target_variables = None
    filter_expr_by_label = None
    period = None
    messages = list()
    survey_scenario = None

    def __init__(self):
        name = self.get_name()
        assert name is not None and isinstance(name, str)

        figures_directory = Path(config.get("paths", "figures_directory")) / name

        if not figures_directory.exists():
            figures_directory.mkdir(parents = True, exist_ok = True)

        self.figures_directory = figures_directory

    def compute_aggregates_comparison(self, input_dataframe_by_entity = None):
        pass

    def compute_distibution_comparison(self, input_dataframe_by_entity = None):
        pass

    def get_name(self):
        return self.name + "_" + str(self.period)

    def get_test_dataframes(self, rebuild = False, noindivs = None):
        start_time = datetime.datetime.now()
        if not rebuild:
            return self._load_test_dataframes()

        input_dataframe_by_entity, target_dataframe_by_entity = self.compute_test_dataframes()
        log.debug(f"Test data has been processed in {datetime.datetime.now() - start_time}")
        save_start_time = datetime.datetime.now()
        try:
            self._save_test_dataframes(input_dataframe_by_entity, target_dataframe_by_entity)
            log.debug(f"Test data has been saved in {datetime.datetime.now() - save_start_time}")
        except Exception as e:
            log.debug(f"Test data has not been saved because of {e}")
            pass

        if noindivs is not None:
            selected_idmen_original = list(input_dataframe_by_entity["individu"].query(f"noindiv in {noindivs}").idmen_original.unique())[0]
            menage_query = f"idmen_original == {selected_idmen_original}"
            selected_noindivs = list(input_dataframe_by_entity["individu"].query(menage_query).noindiv.unique())

            input_dataframe_by_entity = {
                "individu": input_dataframe_by_entity["individu"].query(f"noindiv in {selected_noindivs}"),
                "menage": input_dataframe_by_entity["menage"].query(menage_query),
                }
            target_dataframe_by_entity = {
                "individu": target_dataframe_by_entity["individu"].query(f"noindiv in {selected_noindivs}"),
                "menage": target_dataframe_by_entity["menage"].query(menage_query),
                }
        return input_dataframe_by_entity, target_dataframe_by_entity

    def compare(self, browse, load, verbose, debug, target_variables = None, period = None, rebuild = False, summary = False, compute_divergence = False):
        """Compare actual data with openfisca-france-data computation."""
        log.setLevel(level = logging.DEBUG if verbose else logging.WARNING)

        if target_variables is not None and isinstance(target_variables, str):
            target_variables = [target_variables]

        assert (target_variables is None) or isinstance(target_variables, list)

        if target_variables is None:
            target_variables = self.default_target_variables

        self.target_variables = target_variables   

        if period is not None:
            period = int(period)

        backup_directory = PurePath.joinpath(Path(config.get("paths", "backup")))
        backup_directory.mkdir(parents = True, exist_ok = True)
        name = self.name
        backup_path = PurePath.joinpath(backup_directory, f"{name}_backup.h5")

        if load:
            assert Path.exists(backup_path), f"Backup file {backup_path} doesn't exist"
            shown = pd.read_hdf(
                backup_path,
                'result',
                )
            dtale.show(
                shown,
                open_browser = True,
                subprocess = False,
                )

        try:
            start_time = datetime.datetime.now()
            input_dataframe_by_entity, target_dataframe_by_entity = self.get_test_dataframes(rebuild)

            log.debug(f"Test data has been prepared in {datetime.datetime.now() - start_time}")

            self.compute_aggregates_comparison(
                input_dataframe_by_entity = input_dataframe_by_entity,
                )

            self.compute_distibution_comparison(input_dataframe_by_entity = input_dataframe_by_entity)

            if compute_divergence:
                result_by_variable, markdown_section_by_variable, markdown_summary_section_by_variable = self.compute_divergence(
                    input_dataframe_by_entity = None,  # To force load the data_table from hdf file
                    target_dataframe_by_entity = target_dataframe_by_entity,
                    target_variables = target_variables,
                    period = period,
                    summary = summary,
                    )

            # Deal with markdown_section

                self.create_report(markdown_section_by_variable, markdown_summary_section_by_variable)

                if result_by_variable is None:
                    return

                result = pd.concat(result_by_variable, ignore_index = True)
            else:
                self.create_report(None,None)

            log.debug(f"Eveyrthing has been computed in {datetime.datetime.now() - start_time}")
            del input_dataframe_by_entity, target_dataframe_by_entity


            if browse:
                start_browsing_time = datetime.datetime.now()
                result = result.dropna(axis = 1, how = 'all')
                matching_variables = ["noindiv"]
                assert set(matching_variables) <= set(result.columns)
                cols_to_use = result.columns.tolist() + matching_variables
                shown = result

                log.debug(f"Data for browsing has been prepared in {datetime.datetime.now() - start_browsing_time}")

                dtale.show(
                    shown,
                    open_browser = True,
                    subprocess = False,
                    )

                assert backup_directory.exists()
                shown.to_hdf(
                    backup_path,
                    'result',
                    mode = "w",
                    format = "table",
                    )

        except Exception as error:
            if debug:
                print(error)
                pdb.post_mortem(sys.exc_info()[2])
            raise error

    def compute_divergence(self, input_dataframe_by_entity, target_dataframe_by_entity,
            target_variables = None, period = None, summary = False):
        """
        Compare openfisca-france-data computation with data targets.

        Args:
            input_dataframe_by_period (dict): Input data
            target_dataframe_by_period (dict): Targets to macth
            figures_directory (path): Where to store the figures
        """
        figures_directory = self.figures_directory.resolve()
        assert Path.exists(figures_directory)

        if target_variables is None:
            log.info(f"No target variables. Exiting divergence computation.")
            return None, None, None

        data = (
            dict(input_dataframe_by_entity = input_dataframe_by_entity)
            if input_dataframe_by_entity is not None
            else None
            )

        survey_scenario = self.get_survey_scenario(
            data = data
            )

        tax_benefit_system = survey_scenario.tax_benefit_system
        markdown_section_by_variable = dict()
        markdown_summary_section_by_variable = dict()
        stats_by_variable = dict()
        result_by_variable = dict()

        for variable in target_variables:
            if variable == 'noind':
                continue

            entity = tax_benefit_system.variables[variable].entity.key
            target_dataframe = target_dataframe_by_entity[entity]
            assert variable in target_dataframe
            log.debug(f"Testing final only variable: {variable}")
            result, stats = compute_result(
                variable,
                survey_scenario,
                target_dataframe,
                )

            result_by_variable[variable] = result
            variable_distribution_figures_created = create_variable_distribution_figures(
                variable,
                result,
                figures_directory = figures_directory,
                entity_original_id = get_entity_original_id(survey_scenario, variable),
                )
            diff_variable_distribution_figures_created = create_diff_variable_distribution_figures(
                variable,
                result,
                figures_directory = figures_directory,
                entity_original_id = get_entity_original_id(survey_scenario, variable)
                )
            stats_by_variable[variable] = stats

            variable_markdown_section = create_variable_markdown_section(
                variable,
                stats,
                figures_directory,
                )

            if variable_markdown_section is not None:
                markdown_section_by_variable[variable] = variable_markdown_section

            if summary:
                variable_markdown_summary_section = create_variable_markdown_summary_section(
                    variable,
                    stats,
                    figures_directory,
                    )
                if variable_markdown_summary_section is not None:
                    markdown_summary_section_by_variable[variable] = variable_markdown_summary_section

        return result_by_variable, markdown_section_by_variable, markdown_summary_section_by_variable

    def compute_test_dataframes(self):
        NotImplementedError

    def create_report(self, markdown_section_by_variable, markdown_summary_section_by_variable):
        figures_directory = self.figures_directory

        if self.messages:
            messages_markdown_section = """
Filtres appliqués:

""" + "\n".join(f"- {message}" for message in self.messages) + """
"""
        else:
            messages_markdown_section = ""

        table_agregats_markdown = None
        if PurePath.joinpath(figures_directory, "table_agregats.md").exists():
            with open(figures_directory / "table_agregats.md", "r", encoding = 'utf-8') as table_agregats_md_file:
                table_agregats_markdown = table_agregats_md_file.read()

        distribution_comparison_markdown = None
        if PurePath.joinpath(figures_directory, "distribution_comparison_md").exists():
            with open(figures_directory / "distribution_comparison_md", "r", encoding = 'utf-8') as distribution_comparison_md_file:
                distribution_comparison_markdown = distribution_comparison_md_file.read()

        front_sections = [messages_markdown_section, table_agregats_markdown, distribution_comparison_markdown]
        sections_by_filename = {
            "variables": markdown_section_by_variable,
            "summary_variables": markdown_summary_section_by_variable
            }

        for filename, section_by_variable in sections_by_filename.items():
            if section_by_variable is not None:
                markdown_sections = list(filter(
                    lambda x: x is not None,
                    front_sections + list(section_by_variable.values()),
                    ))
            else:
                markdown_sections = list(filter(
                    lambda x: x is not None,front_sections
                    ))
            create_output_files(
                markdown_sections,
                figures_directory,
                filename,
                )

    def filter(self, data_frame):
        for label, filter_expr in self.filter_expr_by_label.items():
            obs_before = data_frame.noind.nunique()
            selection = (data_frame
                .eval("keep = " + filter_expr)
                .groupby("noindiv")["keep"]
                .transform('all')
                )
            data_frame.drop(data_frame.index[~selection], inplace = True)
            obs_after = data_frame.noind.nunique()
            log_message = f"Applying filter '{label}': dropping {obs_before - obs_after}, keeping {obs_after} observations."
            log.info(log_message)
            self.messages.append(log_message + "\n")

    def get_survey_scenario(self, data = None, survey_name = None):

        if self.survey_scenario is not None:
            return self.survey_scenario

        return get_survey_scenario(
            year = str(self.period),
            data = data,
            survey_name = f'openfisca_erfs_fpr_{self.period}' if survey_name is None else survey_name,
            )

    def _load_test_dataframes(self, noindivs = None, idents = None):
        name = self.get_name()
        backup_directory = PurePath.joinpath(Path(config.get("paths", "backup")))
        backup_path = PurePath.joinpath(backup_directory, f"{name}_test_data.h5")
        assert backup_path.exists(), "Backup data does not exist. Try rebuild option"
        store = pd.HDFStore(backup_path)
        keys = store.keys()
        store.close()
        input_dataframe_by_entity = dict()
        target_dataframe_by_entity = dict()

        entities = ["individu", "menage"]
        for prefix in ["input", "target"]:
            for entity in entities:
                df = pd.read_hdf(backup_path, f'{prefix}_{name}_{entity}')

                if prefix ==  "input":
                    input_dataframe_by_entity[entity] = df
                elif prefix ==  "target":
                    target_dataframe_by_entity[entity] = df

        return input_dataframe_by_entity, target_dataframe_by_entity

    def _save_test_dataframes(self, input_dataframe_by_entity, target_dataframe_by_entity):
        name = self.get_name()
        backup_directory = PurePath.joinpath(Path(config.get("paths", "backup")))
        backup_path = PurePath.joinpath(backup_directory, f"{name}_test_data.h5")
        backup_path.unlink(missing_ok = True)

        for data_frame_by_entity in [input_dataframe_by_entity, target_dataframe_by_entity]:
            if data_frame_by_entity is None:
                continue
            prefix = "target" if id(data_frame_by_entity) == id(target_dataframe_by_entity) else "input"
            for entity, dataframe in data_frame_by_entity.items():
                if dataframe is None:
                    continue
                try:
                    dataframe.to_hdf(
                        backup_path,
                        f'{prefix}_{name}_{entity}',
                        )
                except NotImplementedError:
                    dataframe.to_hdf(
                        backup_path,
                        f'{prefix}_{name}_{entity}',
                        format = "table"
                        )



from openfisca_france_data.erfs_fpr.get_survey_scenario import get_survey_scenario
from openfisca_france_data.reforms.inversion_directe_salaires import inversion_directe_salaires

survey_scenario = get_survey_scenario(year = 2014, reform = inversion_directe_salaires)

data_frame_by_entity = survey_scenario.create_data_frame_by_entity(
    variables = [
        'af_base',
        'af_nbenf',
        'age_en_mois',
        'age',
        'autonomie_financiere',
        'af_eligibilite_base',
        'menage_ordinaire_familles',
        'est_enfant_dans_famille',
        'prestations_familiales_enfant_a_charge',
        'rempli_obligation_scolaire',
        'residence_dom',
        'weight_familles',
        'weight_individus',
        'wprm',
        ],
    )
famille = data_frame_by_entity['famille']
individu = data_frame_by_entity['individu']
menage = data_frame_by_entity['menage']

# %%
population_by_age = individu.groupby('age')['weight_individus'].sum().reset_index()
# Les 0-16 ans sont au moins 700 000 et au plus 850 000
assert (population_by_age.query('age <= 16 & age >= 0')['weight_individus'] > 7e5).all(), \
    'Les 0-16 ans doivent être au moins 700 000'
assert (population_by_age.query('age <= 16 & age >= 0')['weight_individus'] < 85e4).all(), \
    'Les 0-16 ans doivent être sont au plus 850 000'


#%%

assert famille.weight_familles.sum() < 30e6
assert famille.weight_familles.sum() > 29e6

(famille
    .groupby(['af_nbenf'])['weight_familles']
    .sum()
    .reset_index()
    .query('af_nbenf > 0')
    .weight_familles
    .sum()
    ) > 8e6

#%%
print((famille.af_base * famille.weight_familles).sum())


#%%
famille.groupby(['af_eligibilite_base'])['weight_familles'].sum()
menage.groupby(['residence_dom'])['wprm'].sum()
individu.groupby(['prestations_familiales_enfant_a_charge'])['weight_individus'].sum()  # PROBLEM
individu.groupby(['est_enfant_dans_famille'])['weight_individus'].sum()
individu.groupby(['autonomie_financiere'])['weight_individus'].sum()
individu.groupby(
    ['autonomie_financiere', 'est_enfant_dans_famille', 'rempli_obligation_scolaire']
    )['weight_individus'].sum()

#%%
survey_scenario.memory_usage('rempli_obligation_scolaire')

#%%
# survey_scenario.summarize_variable('age')
famille.groupby(['af_nbenf'])['weight_familles'].sum()

#%%
individu.groupby(['age', 'autonomie_financiere'])['weight_individus'].sum()
individu.groupby(['age', 'est_enfant_dans_famille'])['weight_individus'].sum()



import logging
import pandas

from typing import Optional

from openfisca_core import periods  # type: ignore
from openfisca_core.model_api import Enum  # type: ignore
from openfisca_core.taxbenefitsystems import TaxBenefitSystem  # type: ignore
from openfisca_france_data import base_survey as base  # type: ignore
from openfisca_survey_manager.scenarios import AbstractSurveyScenario  # type: ignore

log = logging.getLogger(__name__)


class AbstractErfsSurveyScenario(AbstractSurveyScenario):
    """
    Parties communes entre ERFS et ERFS PFR

    Note : beaucoup de mix entre milléssime, à nettoyer à un moment donné
    """

    period = None  # TODO: déplacer cela dans AbstractSurveyScenario ? C'est un fix suite à https://github.com/openfisca/openfisca-survey-manager/commit/f30ef99fc9f4536406e593693af914516650f458

    id_variable_by_entity_key = dict(
        famille = "idfam",
        foyer_fiscal = "idfoy",
        menage = "idmen",
        )

    filtering_variable_by_entity = dict(
        famille = "menage_ordinaire_familles",
        foyer_fiscal = "menage_ordinaire_foyers_fiscaux",
        individu = "menage_ordinaire_individus",
        menage = "menage_ordinaire",
        )

    role_variable_by_entity_key = dict(
        famille = "quifam",
        foyer_fiscal = "quifoy",
        menage = "quimen",
        )

    weight_variable_by_entity = dict(
        menage = "wprm",
        famille = "weight_familles",
        foyer_fiscal = "weight_foyers",
        individu = "weight_individus",
        )

    def build_input_data_from_test_case(self, test_case_scenario):
        for axe in test_case_scenario["axes"][0]:
            axe["name"] = "salaire_imposable_pour_inversion"

        tax_benefit_system = self.tax_benefit_system

        simulation = (
            tax_benefit_system
            .new_scenario()
            .init_single_entity(**test_case_scenario)
            .new_simulation()
            )

        array_by_variable = dict()
        period = periods.period(str(self.year))

        for variable in self.used_as_input_variables:
            array_by_variable[variable] = simulation.calculate_add(
                variable,
                period = period,
                )

        for ident in ["idmen", "idfoy", "idfam"]:
            array_by_variable[ident] = range(axe["count"])

        input_data_frame = pandas.DataFrame(array_by_variable)

        for qui in ["quimen", "quifoy", "quifam"]:
            input_data_frame[qui] = 0

        return input_data_frame

    @classmethod
    def create(
            cls,
            tax_benefit_system: Optional[TaxBenefitSystem] = None,
            baseline_tax_benefit_system: Optional[TaxBenefitSystem] = None,
            input_data_type = None,
            reform = None,
            reform_key = None,
            year: int = None,
            ):

        assert year is not None
        assert not ((reform is not None) and (reform_key is not None))

        reform_is_provided = (reform is not None) or (reform_key is not None)
        # With booleans != is xor
        # See https://stackoverflow.com/questions/432842/how-do-you-get-the-logical-xor-of-two-variables-in-python # noqa: E501
        assert reform_is_provided != (tax_benefit_system is not None)

        if reform_is_provided:
            assert baseline_tax_benefit_system is not None

        if reform_key is not None:
            reform = base.get_cached_reform(
                reform_key = reform_key,
                tax_benefit_system = baseline_tax_benefit_system,
                )

        if reform is not None:
            tax_benefit_system = reform

        if input_data_type is not None:
            survey_scenario = cls(input_data_type = input_data_type, year = year)

        else:
            survey_scenario = cls(year = year)

        survey_scenario.set_tax_benefit_systems(
            tax_benefit_system = tax_benefit_system,
            baseline_tax_benefit_system = baseline_tax_benefit_system,
            )

        survey_scenario.year = year
        survey_scenario.period = year

        return survey_scenario

    def init_from_test_case(self, test_case_scenario = None):
        assert test_case_scenario is not None
        input_data_frame = self.build_input_data_from_test_case(test_case_scenario)
        self.init_from_data_frame(input_data_frame = input_data_frame)
        self.new_simulation()

        if self.baseline_tax_benefit_system is not None:
            self.new_simulation(use_baseline = True)

    def custom_initialize(self, simulation):
        input_variables = [
            "categorie_salarie",
            "contrat_de_travail",
            "effectif_entreprise",
            "heures_remunerees_volume",
            # 'hsup',
            "pensions_alimentaires_percues",
            ]

        computed_variables_used_as_input =  [
            "chomage_imposable",
            "primes_fonction_publique",
            "retraite_brute",
            "retraite_imposable",
            "salaire_de_base",
            "traitement_indiciaire_brut",
            # 'chomage_brut',
            ]

        three_year_span_variables = input_variables + computed_variables_used_as_input

        simulation_period = periods.period(self.year)
        for variable in three_year_span_variables:
            assert variable in self.used_as_input_variables, \
                f"{variable} is not a in the input_varaibles to be used {self.used_as_input_variables}"  # noqa: E501

            if self.tax_benefit_system.variables[variable].value_type == Enum:
                permanent_value = simulation.calculate(
                    variable,
                    period = periods.period(self.year).first_month,
                    )
            else:
                permanent_value = simulation.calculate_add(
                    variable,
                    period = self.year,
                    )

            for offset in [-1, -2]:
                try:
                    simulation.set_input(
                        variable,
                        simulation_period.offset(offset),
                        permanent_value
                        )
                except ValueError as e:
                    log.debug(f"Dealing with: {e}")
                    if sum(simulation.calculate_add(variable, simulation_period.offset(offset))) != sum(permanent_value):
                        use_baseline = self.baseline_simulation == simulation
                        simulation.delete_arrays(variable, simulation_period.offset(offset))
                        simulation.set_input(
                            variable,
                            simulation_period.offset(offset),
                            permanent_value
                            )


    def custom_input_data_frame(self, input_data_frame, **kwargs):
        for variable in ["quifam", "quifoy", "quimen"]:
            if variable in input_data_frame:
                log.debug(input_data_frame[variable].value_counts(dropna = False))



import gc
import logging
import pandas

from openfisca_france_data.utils import (id_formatter, print_id, normalizes_roles_in_entity)
from openfisca_survey_manager.temporary import temporary_store_decorator  # type: ignore
from openfisca_survey_manager.input_dataframe_generator import set_table_in_survey  # type: ignore

log = logging.getLogger(__name__)


@temporary_store_decorator(file_name = 'erfs_fpr')
def create_input_data_frame(temporary_store = None, year = None, export_flattened_df_filepath = None):
    assert temporary_store is not None
    assert year is not None

    individus = temporary_store['individus_{}'.format(year)]
    menages = temporary_store['menages_{}'.format(year)]

    # ici : variables à garder
    var_individus = [
        'activite',
        'age',
        'categorie_salarie',
        'categorie_non_salarie',
        'chomage_brut',
        'contrat_de_travail',
        'date_naissance',
        'effectif_entreprise',
        'heures_remunerees_volume',
        'idfam',
        'idfoy',
        'idmen',
        'noindiv',
        'pensions_alimentaires_percues',
        'pensions_invalidite',
        'quifam',
        'quifoy',
        'quimen',
        'rag',
        'retraite_brute',
        'ric',
        'rnc',
        'rpns_imposables',
        'salaire_de_base',
        'statut_marital',
        "primes_fonction_publique",
        "traitement_indiciaire_brut",
        ]

    if year >= 2018:
        var_menages = [
            'idmen',
            'loyer',
            'statut_occupation_logement',
            'taxe_habitation',
            'wprm',
            'zone_apl',
            'logement_conventionne',
            'prest_precarite_hand' # on récupère la variable de montant de aah / caah pour pouvoir faire une imputation du handicap
        ]
    else:
        var_menages = [
            'idmen',
            'loyer',
            'taxe_habitation',
            'wprm',
            ]

    individus = create_ids_and_roles(individus)
    individus = individus[var_individus].copy()
    gc.collect()

    # This looks like it could have a sizeable impact
    missingvariablesmenages = ["taxe_habitation"]
    for k in missingvariablesmenages:
        menages[k] = 0


    menages = extract_menages_variables(menages)

    # individus = create_collectives_foyer_variables(individus, menages)

    idmens = individus.idmen.unique()
    menages = menages.loc[
        menages.idmen.isin(idmens),
        var_menages
        ].copy()
    survey_name = 'openfisca_erfs_fpr_' + str(year)

    # Formats ids
    individus = format_ids_and_roles(individus)
    menages = menages.rename(columns = {'idmen':'idmen_original'})
    unique_idmen = individus[['idmen','idmen_original']].drop_duplicates()
    assert len(unique_idmen) == len(menages), "Number of idmen should be the same individus and menages tables."

    menages = menages.merge(unique_idmen,
                            how = 'inner',
                            on = 'idmen_original')

    if export_flattened_df_filepath:
        supermerge = individus.merge(
            menages,
            right_index = True,
            left_on = "idmen",
            suffixes = ("", "_x"))
        log.debug(f"Saving to {export_flattened_df_filepath}")
        supermerge.to_hdf(export_flattened_df_filepath, key = "input")

    # Enters the individual table into the openfisca_erfs_fpr collection
    individus = individus.sort_values(by = ['idmen', 'idfoy', 'idfam'])
    log.debug(f"Saving entity 'individu' in collection 'openfisca_erfs_fpr' and survey name '{survey_name}' with set_table_in_survey")
    set_table_in_survey(
        individus,
        entity = "individu",
        period = year,
        collection = "openfisca_erfs_fpr",
        survey_name = survey_name,
        )

    menages = menages.sort_values(by = ['idmen'])
    log.debug(f"Saving entity 'menage' in collection 'openfisca_erfs_fpr' and survey name '{survey_name}' with set_table_in_survey")
    set_table_in_survey(
        menages,
        entity = "menage",
        period = year,
        collection = "openfisca_erfs_fpr",
        survey_name = survey_name,
        )
    log.debug("End of create_input_data_frame")


def create_collectives_foyer_variables(individus, menages):
    menages_revenus_fonciers = menages[['idmen', 'rev_fonciers_bruts']].copy()
    idmens = menages_revenus_fonciers.query('(rev_fonciers_bruts > 0)')['idmen'].tolist()
    # I'm not proud
    # TODO: WDF ?!
    idmens = [_ for _ in idmens if _ != 14024264]
    idmens = [_ for _ in idmens if _ != 15010850]
    idmens = [_ for _ in idmens if _ != 17000919]
    idmens = [_ for _ in idmens if _ != 18046072]
    idmens = [_ for _ in idmens if _ != 8019138]
    menages_multi_foyers = (individus
        .query('idmen in @idmens')[['idmen', 'idfoy', 'age']]  # On ne garde que les ménages avec des revenus fonciers
        .groupby('idmen')
        .filter(lambda x: x.idfoy.nunique() > 1)  # On ne garde que les ménages avec plus d'un foyer fiscal
        .sort_values(['idmen', 'age'], ascending = False)  # On garde les ages élevés en premier
        # .drop_duplicates(['idmen', 'idfoy'])  # On garde les individus les plus agés d'un même foyer fiscal: inutile
        .drop_duplicates(['idmen'])  # On garde le foyer fiscal du plus agés du ménage
        .drop('age', axis = 1)
        .reset_index(drop = True)
        )
    menages_simple_foyer = (individus
        .query('idmen in @idmens')[['idmen', 'idfoy']]  # On ne garde que les ménages avec des revenus fonciers
        .groupby('idmen')
        .filter(lambda x: x.idfoy.nunique() == 1)  # On ne garde que les ménages avec plus d'un foyer fiscal
        .drop_duplicates(['idmen', 'idfoy'])
        .reset_index(drop = True)
        )

    # update idmens, as some households may have dropped out, causing errors
    # this is kind of a dirty fix, better solution would be to fix the dropping out itself
    # for instance: 2017, 17000919 drops out bc. no "chef de famille" which is weird, neet to investigate
    idmens_old = idmens.copy()
    idmens = set(menages_multi_foyers.idmen.to_list()).union(set(menages_simple_foyer.idmen.to_list()))

    dropouts = set(idmens).symmetric_difference(set(idmens_old))
    if len(dropouts) == 0:
        log.info('No households have been dropped. All clear.')
    else:
        log.warning('Some households [{}] have dropped out. You should investigate why this has happened. [{}]'.format(len(dropouts), ','.join(str(e) for e in dropouts)))

    assert set(menages_multi_foyers.idmen.tolist() + menages_simple_foyer.idmen.tolist()) == set(idmens)
    menages_foyers_correspondance = pandas.concat([menages_multi_foyers, menages_simple_foyer], ignore_index = True)
    del menages_multi_foyers, menages_simple_foyer
    foyers_revenus_fonciers = menages_foyers_correspondance.merge(
        menages_revenus_fonciers,
        how = 'inner',
        on = 'idmen'
        )
    assert len(foyers_revenus_fonciers) == len(idmens)
    del menages_foyers_correspondance, menages_revenus_fonciers
    foyers_revenus_fonciers['quifoy'] = 0
    foyers_revenus_fonciers.drop('idmen', axis = 1, inplace = True)
    assert set(foyers_revenus_fonciers.columns) == set(['idfoy', 'rev_fonciers_bruts', 'quifoy'])
    individus = individus.merge(foyers_revenus_fonciers, how = 'outer', on = ['idfoy', 'quifoy'])
    assert set(idmens) == set(individus .query('(rev_fonciers_bruts > 0)')['idmen'].tolist())
    individus.rename(columns = {'rev_fonciers_bruts': 'revenu_categoriel_foncier'}, inplace = True)
    return individus


def create_ids_and_roles(individus):
    individus.rename(
        columns = {'ident': 'idmen'},
        inplace = True,
        )
    individus['quimen'] = 2

    # checking that only one of "lpr" "lprm" exists, otherwise my
    # great hack wont work
    assert(("lpr" in individus.columns) or ("lprm" in individus.columns))

    lpr = "lpr" if "lpr" in individus.columns else "lprm"
    individus.loc[individus[lpr] == 1, 'quimen'] = 0
    individus.loc[individus[lpr] == 2, 'quimen'] = 1
    individus['idfoy'] = individus['idfam'].copy()
    individus['quifoy'] = individus['quifam'].copy()
    return individus


def format_ids_and_roles(data_frame):
    for entity_id in ['idmen', 'idfoy', 'idfam']:
        data_frame = data_frame.sort_values(by = ['idmen', 'idfoy', 'idfam'])
        log.debug('Reformat ids: {}'.format(entity_id))
        data_frame = id_formatter(data_frame, entity_id)
    data_frame.reset_index(drop = True, inplace = True)
    normalizes_roles_in_entity(data_frame, 'idfoy', 'quifoy')
    normalizes_roles_in_entity(data_frame, 'idmen', 'quimen')
    normalizes_roles_in_entity(data_frame, 'idfam', 'quifam')
    print_id(data_frame)
    return data_frame


@temporary_store_decorator(file_name = 'erfs_fpr')
def extract_menages_variables_from_store(temporary_store = None, year = None):
    assert temporary_store is not None
    assert year is not None
    menages = temporary_store['menages_{}'.format(year)]
    return extract_menages_variables(menages)


def extract_menages_variables(menages):
    variables = ['ident', 'wprm', 'taxe_habitation']
    external_variables = ['loyer', 'zone_apl', 'statut_occupation_logement','logement_conventionne', 'prest_precarite_hand']
    for external_variable in external_variables:
        if external_variable in menages.columns:
            log.debug("Found {} in menages table: we keep it".format(external_variable))
            variables.append(external_variable)
    menages = menages[variables].copy()
    menages.taxe_habitation = - menages.taxe_habitation  # taxes should be negative
    menages.rename(columns = dict(ident = 'idmen'), inplace = True)
    return menages


if __name__ == '__main__':
    import sys
    logging.basicConfig(level = logging.INFO, stream = sys.stdout)
    year = 2014
    data_frame = create_input_data_frame(year = year)

# TODO
# Variables revenus collectifs
#   rev_etranger  -> etr
#   rev_financier_prelev_lib_imputes revenus financiers(prélévement libératoire) et imputés
#   rev_fonciers_bruts Revenus fonciers avant déduction de la CSG -> f4ba  DONE (sur plus agé déclarant du ménage)
#   rev_valeurs_mobilieres_bruts
# Mais aussi:
#   pens_alim_versee
#   Pensions alimentaires versées à des enfants majeurs:
#     décision de justice définitive avant 2006 6GI 6GJ (1er et 2e enfant)
#   Autres pensions alimentaires versées à des enfants majeurs:
#     6EL 6EM
#   Autres pensions alimentaires versées (enfants mineurs, ascendants,…):
#     décision de justice définitive avant 2006  6GP
#   Autres pensions alimentaires versées (enfants mineurs, ascendants,…):
#     6GU
#  th -> taxe_habitation DONE



import logging

from openfisca_france_data import base_survey
from openfisca_france_data.erfs_fpr.get_survey_scenario import get_survey_scenario

log = logging.getLogger(__name__)


def test_rebuild_input_data(year = 2014):

    tax_benefit_system = base_survey.france_data_tax_benefit_system

    survey_scenario = get_survey_scenario(
        tax_benefit_system = tax_benefit_system,
        baseline_tax_benefit_system = tax_benefit_system,
        year = year,
        rebuild_input_data = True,
        )
    return survey_scenario


if __name__ == '__main__':
    import logging
    log = logging.getLogger(__name__)
    import sys
    logging.basicConfig(
        level = logging.DEBUG,
        filename = "rebuild_input_data_log.txt",
        # stream = sys.stdout
        )
    survey_scenario = test_rebuild_input_data()
    survey_scenario.summarize_variable('salaire_de_base', force_compute = True)
    survey_scenario.summarize_variable('salaire_imposable', force_compute = True)



import logging

from openfisca_core.model_api import *  # type:ignore
from openfisca_france.entities import Famille, Individu  # type:ignore
from openfisca_france_data.erfs_fpr.scenario import ErfsFprSurveyScenario  # type:ignore
from openfisca_france_data import france_data_tax_benefit_system  # type:ignore

log = logging.getLogger(__name__)


def get_custom_survey_scenario(year = 2012, rebuild_input_data = False):
    tax_benefit_system = france_data_tax_benefit_system

    class rsa_origin(Variable):
        value_type = float
        entity = Famille
        label = "RSA de l'ENA"
        definition_period = MONTH

    class aah_origin(Variable):
        value_type = float
        entity = Famille
        label = "RSA de l'ENA"
        definition_period = MONTH

    class paje_clca_origin(Variable):
        value_type = float
        entity = Famille
        label = "RSA de l'ENA"
        definition_period = MONTH

    class hors_match(Variable):
        value_type = bool
        entity = Individu
        label = "hors match"
        definition_period = MONTH

    class activite_famille_max(Variable):
        value_type = int
        entity = Famille
        label = "activite_famille"
        definition_period = MONTH

        def formula(famille, period):
            activite = famille.members('activite', period)
            return famille.max(activite, role = Famille.PARENT)

    class activite_famille_min(Variable):
        value_type = int
        entity = Famille
        label = "activite_famille"
        definition_period = MONTH

        def formula(famille, period):
            activite = famille.members('activite', period)
            return famille.min(activite, role = Famille.PARENT)

    tax_benefit_system.add_variable(activite_famille_max)
    tax_benefit_system.add_variable(activite_famille_min)

    survey_scenario = ErfsFprSurveyScenario.create(
        tax_benefit_system = tax_benefit_system,
        year = year,
        )
    data = dict(
        input_data_table = dict(),
        input_data_survey_prefix = 'openfisca_erfs_fpr_data',
        )
    survey_scenario.init_from_data(
        data = data,
        rebuild_input_data = rebuild_input_data,
        )
    return survey_scenario


survey_scenario = get_custom_survey_scenario(rebuild_input_data = False)

data_frame_by_entity = survey_scenario.create_data_frame_by_entity(
    variables = [
        'activite_famille_min',
        'activite_famille_max',
        'aspa',
        'asi',
        'etudiant',
        'hors_match',
        'activite',
        'heures_remunerees_volume',
        'nb_parents',
        'rsa',
        'rsa_base_ressources',
        'rsa_base_ressources_individu',
        'rsa_forfait_logement',
        'rsa_montant',
        'rsa_origin',
        'rsa_revenu_activite',
        'rsa_forfait_logement',
        'weight_familles',
        ],
    )

famille = data_frame_by_entity['famille']
individu = data_frame_by_entity['individu']
menage = data_frame_by_entity['menage']

famille.activite_famille_min.value_counts(dropna = False)

famille.groupby(
    [
        'nb_parents',
        'activite_famille_max',
        'activite_famille_min',
        ],
    )['rsa'].sum()

rsa_pivot_table = survey_scenario.compute_pivot_table(
    columns = ['nb_parents'],
    values = ['rsa'],
    aggfunc = 'sum',
    period = 2012,
    )

count_pivot_table = survey_scenario.compute_pivot_table(
    columns = ['nb_parents'],
    values = ['rsa'],
    aggfunc = 'count',
    period = 2012,
    )

survey_scenario.summarize_variable('rsa', weighted = True, force_compute = True)

survey_scenario.summarize_variable('rsa_revenu_activite')
survey_scenario.summarize_variable('rsa_forfait_logement')
survey_scenario.summarize_variable('rsa_base_ressources')

survey_scenario.summarize_variable('rsa_base_ressources_individu', weighted = True)
survey_scenario.summarize_variable('rsa', weighted = True)

survey_scenario.summarize_variable('revenus_capitaux_prelevement_bareme', weighted = True)
survey_scenario.summarize_variable('revenus_capitaux_prelevement_liberatoire', weighted = True)

types_revenus_non_pros = [
    'allocation_securisation_professionnelle',
    'dedommagement_victime_amiante',
    'gains_exceptionnels',
    'pensions_alimentaires_percues',
    'pensions_invalidite',
    'prestation_compensatoire',
    'prime_forfaitaire_mensuelle_reprise_activite',
    'rsa_base_ressources_patrimoine_individu',
    'rsa_indemnites_journalieres_hors_activite',
    ]

for revenu in types_revenus_non_pros:
    survey_scenario.summarize_variable(revenu, weighted = True)

assert 9e9 < (famille.rsa_montant * famille.weight_familles).sum() / 1e9 < 10e9, \
    "Rsa = {} Mds €".format((famille.rsa_montant * famille.weight_familles).sum() / 1e9)
